__version__ = '0.6.3'

from . import data, models, criterions, ddp, launcher, metrics, parser, utils
from .utils import infoLogger
from freerec.dict2obj import Config


def declare(*, version: str):
    """
    This function checks whether the provided version matches the current version of FreeRec package. 
    If they do not match, a warning message would be printed.
    """
    if version != __version__:
        print(f"\033[1;31m[Warning] FreeRec version of {version} is required but current version is {__version__} \033[0m")




from ..base import RecDataSet


__all__ = []


#TODO:===============================Context-aware Recommendation===============================
class ContextAwareRecSet(RecDataSet):
    DATATYPE =  "Context"




from typing import Tuple, List

import numpy as np
import torchdata.datapipes as dp
from functools import lru_cache

from ..base import RecDataSet
from ...tags import TIMESTAMP, SESSION, ITEM, ID
from ...fields import Field, SparseField, DenseField
from ....utils import infoLogger
from ....dict2obj import Config


__all__ = ['SessionBasedRecSet', 'SessionItemTimeTriplet']


class SessionBasedRecSet(RecDataSet):
    DATATYPE =  "Session"

    def check(self):
        assert isinstance(self.fields[TIMESTAMP], Field), "SessionRecSet must have `TIMESTAMP' field."

    def to_pairs(self, master: Tuple = (SESSION, ID)) -> List:
        return super().to_pairs(master)

    def to_seqs(self, master: Tuple = (SESSION, ID), keepid: bool = False) -> List:
        return super().to_seqs(master, keepid)

    def to_roll_seqs(self, master: Tuple = (SESSION, ID), minlen: int = 2) -> List:
        return super().to_roll_seqs(master, minlen)

    def seqlens(self, master: Tuple = (SESSION, ID)) -> List:
        return super().seqlens(master)

    @lru_cache()
    def has_duplicates(self, master: Tuple = (SESSION, ID)) -> bool:
        from itertools import chain
        train_seqs = self.train().to_seqs(master, keepid=False)
        valid_seqs = self.valid().to_seqs(master, keepid=False)
        test_seqs = self.test().to_seqs(master, keepid=False)
        seqs = map(
            lambda triple: chain(*triple),
            zip(train_seqs, valid_seqs, test_seqs)
        )
        for seq in seqs:
            seq = list(seq)
            if len(seq) != len(set(seq)):
                return True
        return False


class SessionItemTimeTriplet(SessionBasedRecSet):

    VALID_IS_TEST = False

    _cfg = Config(
        fields = [
            (SparseField, Config(name='SessionID', na_value=None, dtype=int, tags=[SESSION, ID])),
            (SparseField, Config(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID])),
            (DenseField, Config(name='Timestamp', na_value=None, dtype=float, tags=[TIMESTAMP], transformer='none'))
        ]
    )

    open_kw = Config(mode='rt', delimiter='\t', skip_lines=1)

    def file_filter(self, filename: str):
        return self.mode in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = datapipe.map(self.row_processer)
        return datapipe

    def summary(self):
        super().summary()
        from prettytable import PrettyTable
        Session, Item = self.fields[SESSION, ID], self.fields[ITEM, ID]

        table = PrettyTable(['#Sessions', '#Items', 'Avg.Len', '#Interactions', '#Train', '#Valid', '#Test', 'Density'])
        trainlens =  self.train().seqlens()
        validlens =  self.valid().seqlens()
        testlens =  self.test().seqlens()
        table.add_row([
            Session.count, Item.count,
            np.mean(
                trainlens + validlens + testlens
            ).item(),
            self.trainsize + self.validsize + self.testsize,
            self.trainsize, self.validsize, self.testsize,
            (self.trainsize + self.validsize + self.testsize) / (Session.count * Item.count)
        ])

        infoLogger(table)




from typing import Iterable, Union, Any, Callable, List, Tuple

import numpy as np
import torchdata.datapipes as dp


__all__ = [
    "DropEmpty", 
    "LeftPruningRow", "RightPruningRow",
    "DropingDuplicates",
    "AddingRow",
    "LeftPaddingRow", "RightPaddingRow",
]


def _to_array(x: Union[np.ndarray, Iterable]) -> np.ndarray:
    return x if isinstance(x, np.ndarray) else np.array(x)

#==================================Filter==================================
class RowFilter(dp.iter.IterDataPipe):
    r"""
    Apply a function to specific indices of each row in an IterableDataset.

    Parameters:
    -----------
    source_dp : tud.IterableDataset
        The source IterableDataset.
    fn : Callable
        The function to apply to the specified indices of each row.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.

    Raises:
    -------
    TypeError
        If `indices` is not an iterable.

    Attributes:
    -----------
    source : tud.IterableDataset
        The source IterableDataset.
    fn : Callable
        The function to apply to the specified indices of each row.
    indices : list of int
        The indices of the elements in each row to which `fn` should be applied.
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        fn: Callable, indices: Iterable[int]
    ):
        super().__init__()

        assert isinstance(indices, Iterable), \
            f"{self.__class__.__name__} requires iterable indices but {type(indices)} recevied ..."

        self.source = source_dp
        self.fn = fn
        self.indices = sorted(set(indices))

    def _apply_fn(self, row: Union[List, Tuple]):
        r"""
        Apply the specified function to the elements of the row at the specified indices.

        Parameters:
        -----------
        row : list or tuple
            The row of data to check.

        Returns:
        --------
        bool
            The final results over specified indices.
        """
        return all(self.fn(row[i]) for i in self.indices)

    def __iter__(self):
        for row in self.source:
            if self._apply_fn(row):
                yield row


@dp.functional_datapipe("drop_empty_")
class DropEmpty(RowFilter):
    r"""
    A functional datapipe that drops empty data from a given datapipe.

    Parameters:
    -----------
    source_dp: IterDataPipe 
        The input datapipe to filter.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    """
    def __init__(
        self, 
        source_dp: dp.iter.IterDataPipe, 
        indices: Iterable[int]
    ) -> None:

        super().__init__(
            source_dp=source_dp,
            fn=self._check,
            indices=indices
        )

    def _check(self, x: Any) -> bool:
        return (not isinstance(x, Iterable)) or len(x) > 0


#==================================Mapper==================================
class RowMapper(dp.iter.IterDataPipe):
    r"""
    Apply a function to specific indices of each row in an IterableDataset.

    Parameters:
    -----------
    source_dp : tud.IterableDataset
        The source IterableDataset.
    fn : Callable
        The function to apply to the specified indices of each row.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.

    Raises:
    -------
    TypeError
        If `indices` is not an iterable.

    Attributes:
    -----------
    source : tud.IterableDataset
        The source IterableDataset.
    fn : Callable
        The function to apply to the specified indices of each row.
    indices : list of int
        The indices of the elements in each row to which `fn` should be applied.
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        fn: Callable, indices: Iterable[int]
    ):
        super().__init__()

        assert isinstance(indices, Iterable), \
            f"{self.__class__.__name__} requires iterable indices but {type(indices)} recevied ..."

        self.source = source_dp
        self.fn = fn
        self.indices = sorted(set(indices))

    def _apply_fn(self, row: Union[List, Tuple]) -> List:
        r"""
        Apply the specified function to the elements of the row at the specified indices.

        Parameters:
        -----------
        row : list or tuple
            The row of data to transform.

        Returns:
        --------
        list
            The transformed row.
        """
        row = row if isinstance(row, list) else list(row)
        for i in self.indices:
            row[i] = self.fn(row[i])
        return row

    def __iter__(self):
        for row in self.source:
            yield self._apply_fn(row)


@dp.functional_datapipe("lprune_")
class LeftPruningRow(RowMapper):
    r"""
    A functional datapipe that prunes the left side of a given datapipe to a specified maximum length.

    Parameters:
    -----------
    source_dp: IterDataPipe 
        The input datapipe to prune.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    maxlen: int 
        The maximum length to prune the input data to.
    """
    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        maxlen: int
    ) -> None:

        self.maxlen = maxlen

        super().__init__(
            source_dp=source_dp, 
            fn=self._lprune,
            indices=indices
        )

    def _lprune(self, x: Iterable) -> Iterable:
        return x[-self.maxlen:]


@dp.functional_datapipe("rprune_")
class RightPruningRow(RowMapper):
    r"""
    A functional datapipe that prunes the right side of a given datapipe to a specified maximum length.

    Parameters:
    -----------
    source_dp: IterDataPipe 
        The input datapipe to prune.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    maxlen: int 
        The maximum length to prune the input data to.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        maxlen: int
    ) -> None:

        self.maxlen = maxlen

        super().__init__(
            source_dp=source_dp, 
            fn=self._rprune,
            indices=indices
        )

    def _rprune(self, x: Iterable) -> Iterable:
        return x[:self.maxlen]


@dp.functional_datapipe("drop_duplicates_")
class DropingDuplicates(RowMapper):
    r"""
    A functional datapipe that drop the duplicates.

    Parameters:
    -----------
    source_dp: IterDataPipe 
        The input datapipe to prune.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    ordered: bool
        `True`: keeping order.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        ordered: bool = True
    ) -> None:

        self.ordered = ordered

        super().__init__(
            source_dp=source_dp, 
            fn=self._drop_in_order if self.ordered else self._drop,
            indices=indices
        )

    def _drop(self, x: Union[np.ndarray, Iterable]) -> Union[np.ndarray, List]:
        if isinstance(x, np.ndarray):
            return np.unique(x)
        else:
            return list(set(x))

    def _drop_in_order(self, x: Union[np.ndarray, Iterable]) -> Union[np.ndarray, List]:
        if isinstance(x, np.ndarray):
            return np.unique(x)
        else:
            return sorted(set(x), key=x.index)


@dp.functional_datapipe("add_")
class AddingRow(RowMapper):
    r"""
    Mapper that adds the input data by a specified offset.

    Parameters:
    -----------
    source_dp: dp.iter.IterDataPipe
        Input datapipeline.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    offset: int
        Amount to add the input data by.   
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        offset: int
    ) -> None:

        self.offset = offset

        super().__init__(
            source_dp=source_dp, 
            fn=self._add,
            indices=indices
        )

    def _add(self, x: Union[np.ndarray, Iterable]) -> np.ndarray:
        r"""
        Examples:
        ---------
        >>> x = [1, 2, 3] 
        >>> _add(x) # self.offset = 1
        [2, 3, 4]
        >>> _add(x) # self.offset = -1
        [0, 1, 2]
        """
        return _to_array(x) + self.offset


@dp.functional_datapipe("lpad_")
class LeftPaddingRow(RowMapper):
    r"""
    A functional data pipeline component that left pads sequences to a maximum length.

    Parameters:
    -----------
    source_dp : dp.iter.IterDataPipe
        The source data pipeline to operate on.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    maxlen : int
        The maximum length to pad the sequences to.
    padding_value : int, optional (default=0)
        The value to use for padding.

    Raises:
    -------
    ValueError:
        To pad an empty element.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        maxlen: int, padding_value: int = 0,
    ) -> None:

        self.maxlen = maxlen
        self.padding_value = padding_value

        super().__init__(
            source_dp=source_dp, 
            fn=self._lpad,
            indices=indices
        )

    def _lpad(self, x: Union[np.ndarray, Iterable]) -> np.ndarray:
        x = _to_array(x)
        if len(x) >= self.maxlen:
            return x
        elif len(x) > 0:
            p = np.empty_like(x[0])
            p.fill(self.padding_value)
            p = np.stack(
                [p] * (self.maxlen - len(x)),
                axis=0
            )
            return np.concatenate((p, x), axis=0)
        else:
            raise ValueError("Cannot pad an empty element")


@dp.functional_datapipe("rpad_")
class RightPaddingRow(RowMapper):
    r"""
    A functional data pipeline component that right pads sequences to a maximum length.

    Parameters:
    -----------
    source_dp : dp.iter.IterDataPipe
        The source data pipeline to operate on.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    maxlen : int
        The maximum length to right pad the sequences to.
    padding_value : int, optional (default=0)
        The value to use for padding.

    Raises:
    -------
    ValueError:
        To pad an empty element.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        maxlen: int, padding_value: int = 0,
    ) -> None:

        self.maxlen = maxlen
        self.padding_value = padding_value

        super().__init__(
            source_dp=source_dp, 
            fn=self._rpad,
            indices=indices
        )

    def _rpad(self, x: Union[np.ndarray, Iterable]) -> np.ndarray:
        x = _to_array(x)
        if len(x) >= self.maxlen:
            return x
        elif len(x) > 0:
            p = np.empty_like(x[0])
            p.fill(self.padding_value)
            p = np.stack(
                [p] * (self.maxlen - len(x)),
                axis=0
            )
            return np.concatenate((x, p), axis=0)
        else:
            raise ValueError("Cannot pad an empty element")




import numpy as np
import torchdata.datapipes as dp

from ..base import RecDataSet
from ...tags import TIMESTAMP, USER, ITEM, ID
from ...fields import Field, SparseField, DenseField
from ....utils import infoLogger
from ....dict2obj import Config


__all__ = ['SequentialRecSet']


class SequentialRecSet(RecDataSet):
    DATATYPE =  "Sequential"

    def check(self):
        assert isinstance(self.fields[TIMESTAMP], Field), "SequentialRecSet must have `TIMESTAMP' field."

    # def raw2pickle(self):
    #     """Convert raw data into (ordered) pickle format."""
    #     infoLogger(f"[{self.__class__.__name__}] >>> Convert raw data ({self.mode}) to chunks in pickle format")
    #     userIndex = self.fields.index(USER, ID)
    #     timeIndex = self.fields.index(TIMESTAMP)
    #     data = sorted(self.raw2data(), key=lambda row: (row[userIndex], row[timeIndex]))
    #     datapipe = dp.iter.IterableWrapper(data)

    #     count = 0
    #     for chunk in datapipe.batch(batch_size=self.DEFAULT_CHUNK_SIZE).column_():
    #         for j, field in enumerate(self.fields):
    #             chunk[j] = field.transform(chunk[j])
    #         self.write_pickle(chunk, count)
    #         count += 1
    #     infoLogger(f"[{self.__class__.__name__}] >>> {count} chunks done")


class UserItemTimeTriplet(SequentialRecSet):

    VALID_IS_TEST = False

    _cfg = Config(
        fields = [
            (SparseField, Config(name='UserID', na_value=None, dtype=int, tags=[USER, ID])),
            (SparseField, Config(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID])),
            (DenseField, Config(name='Timestamp', na_value=None, dtype=float, tags=[TIMESTAMP], transformer='none'))
        ]
    )

    open_kw = Config(mode='rt', delimiter='\t', skip_lines=1)

    def file_filter(self, filename: str):
        return self.mode in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = datapipe.map(self.row_processer)
        return datapipe

    def summary(self):
        super().summary()
        from prettytable import PrettyTable
        User, Item = self.fields[USER, ID], self.fields[ITEM, ID]

        table = PrettyTable(['#Users', '#Items', 'Avg.Len', '#Interactions', '#Train', '#Valid', '#Test', 'Density'])
        table.add_row([
            User.count, Item.count, self.train().meanlen + 2,
            self.trainsize + self.validsize + self.testsize,
            self.trainsize, self.validsize, self.testsize,
            (self.trainsize + self.validsize + self.testsize) / (User.count * Item.count)
        ])

        infoLogger(table)




from typing import Iterator, Iterable, List, Union, TypeVar, Any, Callable, Optional

import numpy as np
import torch, warnings
import torchdata.datapipes as dp
from functools import partial

from .base import Postprocessor
from ..fields import BufferField, FieldList, Field


__all__ = [
    "Columner", "ToTensor", "Fielder",
    'LeftPaddingCol', 'RightPaddingCol'
]


def _to_array(x: Union[np.ndarray, Iterable]) -> np.ndarray:
    return x if isinstance(x, np.ndarray) else np.array(x)


T = TypeVar("T")


@dp.functional_datapipe("column_")
class Columner(Postprocessor):
    r"""
    A postprocessor that converts a batch of samples into columns.
    Columner takes a datapipe that yields a batch of samples, and converts them into columns. 
    This can be useful for the following transformation (by column).

    Parameters:
    -----------
    source_dp: dp.IterDataPipe 
        A datapipe that yields a batch samples.
    """

    def __iter__(self):
        for batch in self.source:
            yield list(zip(*batch))


@dp.functional_datapipe("tensor_")
class ToTensor(Postprocessor):
    r"""
    A datapipe that converts lists into torch Tensors.
    This class converts a List into a torch.Tensor.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, *, 
        idtype = torch.long, fdtype = torch.float32
    ) -> None:
        super().__init__(source_dp)
        self.idtype = idtype
        self.fdtype = fdtype

    def at_least_2d(self, vals: torch.Tensor):
        """Reshape tensor to 2D if needed."""
        return vals.unsqueeze(1) if vals.ndim == 1 else vals

    def to_tensor(self, col: List) -> Union[List, torch.Tensor]:
        """Convert the List to a torch.Tensor."""
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                arr = np.stack(col)
            if arr.dtype.kind == 'i':
                ter = torch.as_tensor(arr, dtype=self.idtype)
            elif arr.dtype.kind == 'f': 
                ter = torch.as_tensor(arr, dtype=self.fdtype)
            else:
                ter = torch.as_tensor(arr)
            return self.at_least_2d(ter)
        except ValueError: # skip ragged List
            return col
        except TypeError: # skip objects
            return col

    def __iter__(self) -> Iterator:
        for cols in self.source:
            yield self.listmap(self.to_tensor, cols)


@dp.functional_datapipe("field_")
class Fielder(Postprocessor):
    r"""
    Convert column data to field-style data.
    Then filtering by tags can be possible.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        *fields: Field
    ) -> None:
        super().__init__(source_dp, fields=fields)

    @staticmethod
    def _buffer(field: Field, col: Any):
        return field.buffer(col)

    def __iter__(self) -> Iterator[FieldList[BufferField]]:
        for cols in self.source:
            yield FieldList(map(
                self._buffer,
                self.fields,
                cols
            ))


class ColMapper(dp.iter.IterDataPipe):
    r"""
    Apply a function to specific indices of some columns in an IterableDataset.

    Parameters:
    -----------
    source_dp : tud.IterableDataset
        The source IterableDataset.
    fn : Callable
        The function to apply to the specified indices of each row.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.

    Raises:
    -------
    TypeError
        If `indices` is not an iterable.

    Attributes:
    -----------
    source : tud.IterableDataset
        The source IterableDataset.
    fn : Callable
        The function to apply to the specified indices of each row.
    indices : list of int
        The indices of the elements in each row to which `fn` should be applied.
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        fn: Callable, indices: Iterable[int]
    ):
        super().__init__()

        assert isinstance(indices, Iterable), \
            f"{self.__class__.__name__} requires iterable indices but {type(indices)} recevied ..."

        self.source = source_dp
        self.fn = fn
        self.indices = sorted(set(indices))

    def _apply_fn(self, cols: Iterable):
        r"""
        Apply the specified function to the elements of the row at the specified indices.

        Parameters:
        -----------
        cols : Iterable[Union[List, Tuple, Tensor]]
            The col of data to transform.

        Returns:
        --------
        list
            The transformed cols.
        """
        if isinstance(cols, torch.Tensor):
            cols = cols.tolist()
        elif not isinstance(cols, List):
            cols = list(cols)
        for i in self.indices:
            cols[i] = self.fn(cols[i])
        return cols

    def __iter__(self):
        for cols in self.source:
            yield self._apply_fn(cols)

        
@dp.functional_datapipe("lpad_col_")
class LeftPaddingCol(ColMapper):
    r"""
    A functional data pipeline component that left pads sequences to a maximum length.

    Parameters:
    -----------
    source_dp : dp.iter.IterDataPipe
        The source data pipeline to operate on.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    maxlen : int, optional
        The maximum length to pad the sequences to.
        `None`: `maxlen' will be the maximum length of the current batch
    padding_value : int, optional (default=0)
        The value to use for padding.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        maxlen: Optional[int] = None, padding_value: int = 0,
    ) -> None:

        self.maxlen = maxlen
        self.padding_value = padding_value

        super().__init__(
            source_dp=source_dp, 
            fn=self._lpad,
            indices=indices
        )

    def _zero_like(self, x: Union[Iterable, int, float]):
        if isinstance(x, Iterable):
            return [self._zero_like(item) for item in x]
        else:
            return self.padding_value

    def _lpad_row(self, x: Iterable, maxlen: int):
        x = _to_array(x)
        if len(x) >= maxlen:
            return x
        elif len(x) > 0:
            p = np.empty_like(x[0])
            p.fill(self.padding_value)
            p = np.stack(
                [p] * (maxlen - len(x)),
                axis=0
            )
            return np.concatenate((p, x), axis=0)
        else:
            raise ValueError("Cannot pad an empty element")

    def _lpad(self, rows: Iterable):
        rows = rows if isinstance(rows, List) else list(rows)
        maxlen = maxlen if self.maxlen is not None else max([len(row) for row in rows])
        return list(map(
            partial(self._lpad_row, maxlen=maxlen),
            rows
        ))


@dp.functional_datapipe("rpad_col_")
class RightPaddingCol(ColMapper):
    r"""
    A functional data pipeline component that right pads sequences to a maximum length.

    Parameters:
    -----------
    source_dp : dp.iter.IterDataPipe
        The source data pipeline to operate on.
    indices : Iterable[int]
        The indices of the elements in each row to which `fn` should be applied.
    maxlen : int
        The maximum length to pad the sequences to.
        `None`: `maxlen' will be the maximum length of the current batch
    padding_value : int, optional (default=0)
        The value to use for padding.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        indices: Union[None, int, Iterable[int]],
        maxlen: Optional[int] = None, padding_value: int = 0,
    ) -> None:

        self.maxlen = maxlen
        self.padding_value = padding_value

        super().__init__(
            source_dp=source_dp, 
            fn=self._rpad,
            indices=indices
        )

    def _zero_like(self, x: Union[Iterable, int, float]):
        if isinstance(x, Iterable):
            return [self._zero_like(item) for item in x]
        else:
            return self.padding_value

    def _rpad_row(self, x: Iterable, maxlen: int):
        x = _to_array(x)
        if len(x) >= maxlen:
            return x
        elif len(x) > 0:
            p = np.empty_like(x[0])
            p.fill(self.padding_value)
            p = np.stack(
                [p] * (maxlen - len(x)),
                axis=0
            )
            return np.concatenate((x, p), axis=0)
        else:
            raise ValueError("Cannot pad an empty element")

    def _rpad(self, rows: Iterable):
        rows = rows if isinstance(rows, List) else list(rows)
        maxlen = maxlen if self.maxlen is not None else max([len(row) for row in rows])
        return list(map(
            partial(self._rpad_row, maxlen=maxlen),
            rows
        ))


# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""
This file contains primitives for multi-gpu communication.
This is useful when doing distributed training.
"""

from typing import List, TypeVar, Dict

import torch
import numpy as np
import os, pickle, functools, warnings
import torch.distributed as dist


_LOCAL_PROCESS_GROUP = None
"""
A torch process group which only includes processes that on the same machine as the current process.
This variable is set when processes are spawned by `launch()` in "engine/launch.py".
"""

T = TypeVar("T")


def get_world_size() -> int:
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size()
    return 1

def get_rank() -> int:
    if dist.is_available() and dist.is_initialized():
        return dist.get_rank()
    return 0

def get_local_rank() -> int:
    """
    Returns:
        The rank of the current process within the local (per-machine) process group.
    """
    if dist.is_available() and dist.is_initialized():
        assert _LOCAL_PROCESS_GROUP is not None
        return dist.get_rank(group=_LOCAL_PROCESS_GROUP)
    return 0

def get_local_size() -> int:
    """
    Returns:
        The size of the per-machine process group,
        i.e. the number of processes per machine.
    """
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size(group=_LOCAL_PROCESS_GROUP)
    return 1

def is_distributed() -> bool:
    world_size = os.environ.get('WORLD_SIZE', None)
    # world_size > 1 will be better?
    return world_size is not None

def is_main_process() -> bool:
    """Checks if the current process is the main process"""
    return get_rank() == 0

def main_process_only(func):
    def wrapper(*args, **kwargs):
        if is_main_process():
            return func(*args, **kwargs)
    wrapper.__name__ = func.__name__
    wrapper.__func__ = func.__doc__
    return wrapper

def synchronize():
    """
    Helper function to synchronize (barrier) among all processes when
    using distributed training
    """
    if get_world_size() == 1:
        return
    dist.barrier()

@functools.lru_cache()
def _get_global_gloo_group():
    """
    Return a process group based on gloo backend, containing all the ranks
    The result is cached.
    """
    if dist.get_backend() == "nccl":
        return dist.new_group(backend="gloo")
    else:
        return dist.group.WORLD

def _serialize_to_tensor(data, group):
    backend = dist.get_backend(group)
    assert backend in ["gloo", "nccl"]
    device = torch.device("cpu" if backend == "gloo" else "cuda")

    buffer = pickle.dumps(data)
    if len(buffer) > 1024 ** 3:
        warnings.warn(
            "Rank {} trying to all-gather {:.2f} GB of data on device {}".format(
                get_rank(), len(buffer) / (1024 ** 3), device
            )
        )
    storage = torch.ByteStorage.from_buffer(buffer)
    tensor = torch.ByteTensor(storage).to(device=device)
    return tensor

def _pad_to_largest_tensor(tensor, group):
    """
    Returns:
        list[int]: size of the tensor, on each rank
        Tensor: padded tensor that has the max size
    """
    world_size = dist.get_world_size(group=group)
    assert (
        world_size >= 1
    ), "comm.gather/all_gather must be called from ranks within the given group!"
    local_size = torch.tensor(
        [tensor.numel()], dtype=torch.int64, device=tensor.device)
    size_list = [
        torch.zeros([1], dtype=torch.int64, device=tensor.device)
        for _ in range(world_size)
    ]
    dist.all_gather(size_list, local_size, group=group)
    size_list = [int(size.item()) for size in size_list]

    max_size = max(size_list)

    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    if local_size != max_size:
        padding = torch.zeros(
            (max_size - local_size,), dtype=torch.uint8, device=tensor.device
        )
        tensor = torch.cat((tensor, padding), dim=0)
    return size_list, tensor

def all_gather(data: T, group=None) -> List[T]:
    r"""
    Run all_gather on arbitrary picklable data (not necessarily tensors).
    
    Parameters:
    -----------
    data: any picklable object
    group: a torch process group. By default, will use a group which
        contains all ranks on gloo backend.

    Returns:
    --------
        list[data]: list of data gathered from each rank
    """
    if get_world_size() == 1:
        return [data]

    device = None
    if isinstance(data, torch.Tensor):
        device = data.device
        data = data.cpu()
    if group is None:
        group = _get_global_gloo_group()
    if dist.get_world_size(group) == 1:
        return [data]

    tensor = _serialize_to_tensor(data, group)

    size_list, tensor = _pad_to_largest_tensor(tensor, group)
    max_size = max(size_list)

    # receiving Tensor from all ranks
    tensor_list = [
        torch.empty((max_size,), dtype=torch.uint8, device=tensor.device)
        for _ in size_list
    ]
    dist.all_gather(tensor_list, tensor, group=group)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    if device:
        data_list = [data.to(device) for data in data_list]
    return data_list

def gather(data: T, dst: int = 0, group=None) -> List[T]:
    r"""
    Run gather on arbitrary picklable data (not necessarily tensors).

    Parameters:
    -----------
    data: any picklable object
    dst (int): destination rank
    group: a torch process group. By default, will use a group which
        contains all ranks on gloo backend.

    Returns:
    --------
    list[data]: on dst, a list of data gathered from each rank. Otherwise,
        an empty list.
    """
    if get_world_size() == 1:
        return [data]

    device = None
    if isinstance(data, torch.Tensor):
        device = data.device
        data = data.cpu()
    if group is None:
        group = _get_global_gloo_group()
    if dist.get_world_size(group=group) == 1:
        return [data]
    rank = dist.get_rank(group=group)

    tensor = _serialize_to_tensor(data, group)
    device = tensor.device
    size_list, tensor = _pad_to_largest_tensor(tensor, group)

    # receiving Tensor from all ranks
    if rank == dst:
        max_size = max(size_list)
        tensor_list = [
            torch.empty((max_size,), dtype=torch.uint8, device=device)
            for _ in size_list
        ]
        dist.gather(tensor, tensor_list, dst=dst, group=group)

        data_list = []
        for size, tensor in zip(size_list, tensor_list):
            buffer = tensor.cpu().numpy().tobytes()[:size]
            data_list.append(pickle.loads(buffer))
        if device:
            data_list = [data.to(device) for data in data_list]
        return data_list
    else:
        dist.gather(tensor, [], dst=dst, group=group)
        return []

def shared_random_seed() -> int:
    r"""
    Returns:
    --------
    int: a random number that is the same across all workers.
        If workers need a shared RNG, they can use this shared seed to
        create one.

    Notes:
    ------
    All workers must call this function, otherwise it will deadlock.
    """
    ints = np.random.randint(2 ** 31)
    all_ints = all_gather(ints)
    return all_ints[0]

def reduce_dict(input_dict: Dict, average: bool = True) -> Dict:
    r"""
    Reduce the values in the dictionary from all processes so that process with rank
    0 has the reduced results.

    Parameters:
    -----------
    input_dict: dict
        inputs to be reduced. (values not necessarily tensors).
    average: bool, default to `True`
        whether to do average or sum

    Returns:
    --------
        a dict with the same keys as input_dict, after reduction.
    """

    world_size = get_world_size()
    if world_size < 2:
        return input_dict

    with torch.no_grad():

        # Convert to CUDA Tensor for dist.reduce()
        input_dict_cuda_vals = {}
        for k, v in input_dict.items():
            if type(v) == torch.Tensor:
                input_dict_cuda_vals[k] = v.to('cuda')
            else:
                input_dict_cuda_vals[k] = torch.tensor(v, device='cuda')

        names = []
        values = []
        for k, v in sorted(input_dict_cuda_vals.items()):
            names.append(k)
            values.append(v)
        values = torch.stack(values, dim=0)
        dist.reduce(values, dst=0) # reduce to gpu 0

        if dist.get_rank() == 0 and average:
            # only main process gets accumulated, so only divide by
            # world_size in this case
            values /= world_size
        reduced_dict = {k: v for k, v in zip(names, values)}
    return reduced_dict




from typing import Tuple, Iterable, Optional, Union, Mapping

import os, random
import numpy as np
import pandas as pd
from math import floor, ceil

from ..tags import USER, SESSION, ITEM, RATING, TIMESTAMP

from ...utils import infoLogger, warnLogger, mkdirs


__all__ = ['AtomicConverter']


NAME_FORMAT_DICT = {
    'user_id': USER.name,
    'session_id': SESSION.name,
    'item_id': ITEM.name,
    'venue_id': ITEM.name, # FourSquare
    'rating': RATING.name,
    'timestamp': TIMESTAMP.name
}


TYPE_FORMAT_DICT = {
    'token': str,
    'token_seq': str,
    'float': float,
    # 'float_seq': None
}


class AtomicConverter:
    r"""
    How to convert Atomic files into train|valid|test.txt ?

    Flows:
    ------
    1. Download corresponding files from [[RecBole](https://drive.google.com/drive/folders/1so0lckI6N6_niVEYaBu-LIcpOdZf99kj)].
    2. Import corresponding dataset.
    3. Call `make_xxx_dataset'.

    Parameters:
    -----------
    root: str
        The root of data.
    filename: str
        The file with Atomic files.
        - `None': Use cls.filename instead.
    dataset: str, optional
        The filename saving dataset.
        - `None': Use classname instead.

    Notes:
    ------
    1. Most of datasets in RecBole have two versions: 
        merged: remove duplicate interactions.
        not_merged: not remove duplicate interactions.
    """
    filename: str

    _name_format_dict = dict()
    _type_format_dict = dict()

    def __init__(
        self, root: str, 
        filename: Optional[str] = None,
        dataset: Optional[str] = None
    ) -> None:
        super().__init__()

        filename = filename if filename else self.filename
        self.path = os.path.join(root, filename)
        if not os.path.exists(self.path) or not any(True for _ in os.scandir(self.path)):
            raise FileNotFoundError(f"No such file of {self.path}, or this dir is empty ...")

        self.root = root
        self.dataset = dataset if dataset else self.__class__.__name__

        for key, val in NAME_FORMAT_DICT.items():
            if self._name_format_dict.get(key, None) is None:
                self._name_format_dict[key] = val
        for key, val in TYPE_FORMAT_DICT.items():
            if self._type_format_dict.get(key, None) is None:
                self._type_format_dict[key] = val

    def convert_by_column(self, df: pd.DataFrame):
        old_columns = df.columns
        new_columns = []

        for col in old_columns:
            col_ = col.lower()
            name_, type_ = col_.split(":")
            name_ =  self._name_format_dict.get(name_, name_.capitalize())
            type_ = self._type_format_dict[type_]
            try:
                df[col] = df[col].astype(type_)
            except ValueError as e:
                warnLogger(
                    f"`{type_}' cannot address field `{col}': \n"
                    f"\t {e} \n"
                    f"Skip it ..."
                )
            new_columns.append(name_)
        
        df.columns = new_columns
        return df

    def load_inter_file(self):
        for file_ in os.scandir(self.path):
            filename = file_.name
            if not filename.endswith('.inter'):
                continue
            file_ = os.path.join(self.path, filename)
            df = pd.read_csv(file_, delimiter='\t')
            infoLogger(f"[Converter] >>> Load `{filename}' ...")
            return self.convert_by_column(df)
        infoLogger(f"[Converter] >>> No file ends with `.inter' ...")

    def load_user_file(self):
        for file_ in os.scandir(self.path):
            filename = file_.name
            if not filename.endswith('.user'):
                continue
            file_ = os.path.join(self.path, filename)
            df = pd.read_csv(file_, delimiter='\t')
            infoLogger(f"[Converter] >>> Load `{filename}' ...")
            return self.convert_by_column(df)
        infoLogger(f"[Converter] >>> No file ends with `.user' ...")

    def load_item_file(self):
        for file_ in os.scandir(self.path):
            filename = file_.name
            if not filename.endswith('.item'):
                continue
            file_ = os.path.join(self.path, filename)
            df = pd.read_csv(file_, delimiter='\t')
            infoLogger(f"[Converter] >>> Load `{filename}' ...")
            return self.convert_by_column(df)
        infoLogger(f"[Converter] >>> No file ends with `.item' ...")

    def load_kg_file(self):
        raise NotImplementedError()

    def load_link_file(self):
        raise NotImplementedError()

    def load_net_file(self):
        raise NotImplementedError()

    def load(self):
        self.interactions = self.load_inter_file()
        self.userFeats = self.load_user_file()
        self.itemFeats = self.load_item_file()
        self.pools = (self.interactions, self.userFeats, self.itemFeats)

    def reserve(self, columns: Iterable[str]):
        all_columns = self.interactions.columns
        columns_ = set(columns) & set(all_columns)
        columns = sorted(columns_, key=columns.index)
        infoLogger(f"[Converter] >>> Reserve fields: {columns} ...")
        self.interactions = self.interactions[columns]

    def exclude(self, columns: Iterable[str]):
        all_columns = self.interactions.columns
        columns = set(all_columns) - set(columns)
        columns = sorted(columns, key=all_columns.index)
        infoLogger(f"[Converter] >>> Exclude fields: {columns} ...")
        self.interactions = self.interactions[columns]

    def reorder(self, columns: Iterable[str] = (USER.name, ITEM.name)):
        all_columns = self.interactions.columns
        columns = columns + all_columns
        columns_ = set(columns) & set(all_columns)
        columns = sorted(columns_, key=columns.index)
        infoLogger(f"[Converter] >>> Reorder fields: {columns} ...")
        self.interactions = self.interactions[columns]

    def map_col(
        self, col: str, mapper: Mapping, 
        pools: Optional[Iterable[pd.DataFrame]] = None
    ):
        pools = self.pools if pools is None else pools
        for df in pools:
            if df is not None:
                df[col] = df[col].map(mapper)

    def filter_by_rating(
        self, 
        low: Union[None, int, float] = None, 
        high: Union[None, int, float] = None
    ):
        try:
            df = self.interactions
            ratings = df[RATING.name]
            low = low if low is not None else ratings.min()
            high = high if high is not None else ratings.max()
            self.interactions = df[(low <= ratings) & (ratings <= high)]
            infoLogger(f"[Converter] >>> Filter dataframe according to {RATING.name} ...")
        except KeyError:
            infoLogger(
                f"[Converter] >>> Skip `filter_by_rating` because of `inter` has no field of `{RATING.name}' ..."
            )
    
    def filter_by_core(
        self,
        low4user: Union[None, int, float] = None, 
        high4user: Union[None, int, float] = None,
        low4item: Union[None, int, float] = None, 
        high4item: Union[None, int, float] = None,
        master: str = USER.name,
        strict: bool = True
    ):
        r"""
        Filter (user, item) by k-core settings.

        Parameters:
        -----------
        low4user: int, float, optional
            Minimum core for user.
            - `None`: float('-inf')
        max4user: int, float, optional
            Maximum core for user.
            - `None`: float('inf')
        low4item: int, float, optional
            Minimum core for item.
            - `None`: float('-inf')
        max4item: int, float, optional
            Maximum core for item.
            - `None`: float('inf')
        strict: bool, default to `True`
            `True`: strictly filter by core
            `False`: filter by core only once
        """
        low4user = low4user if low4user else float('-inf')
        high4user = high4user if high4user else float('inf')
        low4item = low4item if low4item else float('-inf')
        high4item = high4item if high4item else float('inf')
        df = self.interactions
        dsz = -1
        infoLogger(
            f"[Converter] >>> Filter dataframe: "
            f"{master} in [{low4user}, {high4user}]; "
            f"Item in [{low4item}, {high4item}] ..."
        )
        infoLogger(f"[Converter] >>> Current datasize: {len(df)} ...")
        while dsz != len(df):
            dsz = len(df)

            # filter by user
            users = df[master]
            counts = users.value_counts()
            bool_indices = users.isin(
                counts[(low4user <= counts) & (counts <= high4user)].index
            )
            df = df[bool_indices]

            # filter by item
            items = df[ITEM.name]
            counts = items.value_counts()
            bool_indices = items.isin(
                counts[(low4item <= counts) & (counts <= high4item)].index
            )
            df = df[bool_indices]

            infoLogger(f"[Converter] >>> Current datasize: {len(df)}")

            if not strict:
                break

        self.interactions = df

    def user2token(self, master: str = USER.name):
        infoLogger(f"[Converter] >>> Map user ID to Token ...")
        user_ids = sorted(self.interactions[master].unique().tolist())
        self.userCount = len(user_ids)

        user_tokens = list(range(len(user_ids)))
        self.userMaps = dict(
            zip(user_ids, user_tokens)
        )

        if self.userFeats is not None:
            self.userFeats = self.userFeats[self.userFeats[master].isin(user_ids)]
            self.userFeats = self.userFeats.sort_values([master])

        self.map_col(
            master, self.userMaps, 
            pools=(self.interactions, self.userFeats)
        )

    def item2token(self):
        infoLogger(f"[Converter] >>> Map item ID to Token ...")
        item_ids = sorted(self.interactions[ITEM.name].unique().tolist())
        self.itemCount = len(item_ids)

        item_tokens = range(len(item_ids))
        self.itemMaps = dict(
            zip(item_ids, item_tokens)
        )

        if self.itemFeats is not None:
            self.itemFeats = self.itemFeats[self.itemFeats[ITEM.name].isin(item_ids)]
            self.itemFeats = self.itemFeats.sort_values([ITEM.name])

        self.map_col(
            ITEM.name, self.itemMaps, 
            pools=(self.interactions, self.itemFeats)
        )

    def sort_by_timestamp(self, master: str = USER.name):
        df = self.interactions
        try:
            df = df.sort_values(by=[master, TIMESTAMP.name])
            infoLogger(f"[Converter] >>> Sort by [{master}] [{TIMESTAMP.name}] ...")
        except KeyError:
            df = df.sort_values(by=[master])
            infoLogger(f"[Converter] >>> Sort by [{master}] ...")
        finally:
            self.interactions = df

    def gen_split_by_ratio(self, ratios: Iterable = (8, 1, 1)):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")
        traingroups = []
        validgroups = []
        testgroups = []
        markers = np.cumsum(ratios)
        for _, group in self.interactions.groupby(USER.name):
            if len(group) == 0:
                continue
            l = max(floor(markers[0] * len(group) / markers[-1]), 1)
            r = floor(markers[1] * len(group) / markers[-1])
            traingroups.append(group[:l])
            if l < r:
                validgroups.append(group[l:r])
            if r < len(group):
                testgroups.append(group[r:])

        self.trainiter = pd.concat(traingroups).reset_index(drop=True)
        self.validiter = pd.concat(validgroups).reset_index(drop=True)
        self.testiter = pd.concat(testgroups).reset_index(drop=True)

    def seq_split_by_ratio(self, ratios: Iterable = (8, 1, 1), seed: int = 0):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")
        markers = np.cumsum(ratios)
        groups = [pair[1] for pair in self.interactions.groupby(USER.name)]

        l = max(floor(markers[0] * len(groups) / markers[-1]), 1)
        r = floor(markers[1] * len(groups) / markers[-1])

        random.seed(seed)
        random.shuffle(groups)

        traingroups = groups[:l]
        validgroups = groups[l:r]
        testgroups = groups[r:]

        self.trainiter = pd.concat(traingroups).reset_index(drop=True)
        self.validiter = pd.concat(validgroups).reset_index(drop=True)
        self.testiter = pd.concat(testgroups).reset_index(drop=True)

    def seq_split_by_last_two(self):
        infoLogger(f"[Converter] >>> Split by leaving last two ...")
        traingroups = []
        validgroups = []
        testgroups = []
        for _, group in self.interactions.groupby(USER.name):
            if len(group) == 0:
                continue
            if len(group) <= 3:
                # Note that sequence with len(group) == 3 cannot be splited into train/valid/test,
                # because train sequence needs at least length >= 2 for prediction.
                traingroups.append(group)
            else:
                traingroups.append(group[:-2])
                validgroups.append(group[-2:-1])
                testgroups.append(group[-1:])

        self.trainiter = pd.concat(traingroups).reset_index(drop=True)
        self.validiter = pd.concat(validgroups).reset_index(drop=True)
        self.testiter = pd.concat(testgroups).reset_index(drop=True)

    def sess_split_by_ratio(self, ratios: Iterable = (8, 1, 1)):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")

        # max(): choosing the last timestamp as the timestamp for the session
        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)

        markers = np.cumsum(ratios)
        l = max(floor(markers[0] * len(groups) / markers[-1]), 1)
        r = floor(markers[1] * len(groups) / markers[-1])

        traingroups = groups[:l]
        validgroups = groups[l:r]
        testgroups = groups[r:]

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]

    def sess_split_by_day(self, days: int = 1):
        infoLogger(f"[Converter] >>> Split by days: {days} ...")

        seconds_per_day = 86400
        seconds = seconds_per_day * days
        last_date = self.interactions[TIMESTAMP.name].max().item()

        # Group interactions by session and calculate session timestamps
        session_timestamps = self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values()

        # Split interactions into train, validation, and test sets based on session timestamps
        traingroups = session_timestamps[session_timestamps < (last_date - 2 * seconds)].index
        validgroups = session_timestamps[(session_timestamps >= (last_date - 2 * seconds)) & (session_timestamps < (last_date - seconds))].index
        testgroups = session_timestamps[session_timestamps >= (last_date - seconds)].index

        assert len(traingroups) >= 0, f"The given `days` of {days} leads to zero-size trainsets ..."
        assert len(validgroups) >= 0, f"The given `days` of {days} leads to zero-size validsets ..."
        assert len(testgroups) >= 0, f"The given `days` of {days} leads to zero-size testsets ..."

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]

    def save(self, path: str):
        mkdirs(path)

        infoLogger(f"[Converter] >>> Save `train.txt' to {path} ...")
        df = pd.DataFrame(self.trainiter)
        df.to_csv(os.path.join(path, 'train.txt'), sep='\t', index=False)

        infoLogger(f"[Converter] >>> Save `valid.txt' to {path} ...")
        df = pd.DataFrame(self.validiter)
        df.to_csv(os.path.join(path, 'valid.txt'), sep='\t', index=False)

        infoLogger(f"[Converter] >>> Save `test.txt' to {path} ...")
        df = pd.DataFrame(self.testiter)
        df.to_csv(os.path.join(path, 'test.txt'), sep='\t', index=False)

        if self.userFeats is not None:
            infoLogger(f"[Converter] >>> Save `user.txt' to {path} ...")
            self.userFeats.to_csv(
                os.path.join(path, 'user.txt'),
                sep='\t', index=False
            )

        if self.itemFeats is not None:
            infoLogger(f"[Converter] >>> Save `item.txt' to {path} ...")
            self.itemFeats.to_csv(
                os.path.join(path, 'item.txt'),
                sep='\t', index=False
            )

        self.summary()

    def make_general_dataset(
        self,
        star4pos: int = 0,
        kcore4user: int = 10,
        kcore4item: int = 10,
        strict: bool = True,
        ratios: Tuple[int, int, int] = (8, 1, 1),
        fields: Optional[Iterable[str]] = (USER.name, ITEM.name)
    ):
        r"""
        Make general dataset.

        Parameters:
        -----------
        star4pos: int, default to 0
            Select interactions with `Rating > star4pos'.
        kcore4user: int, default to 10
            Select kcore interactions according to User.
        kcore4item: int, default to 10
            Select kcore interactions according to Item.
        strict: bool, default to `True`
            `True`: strictly filter by core
            `False`: filter by core only once
        ratios: Tuple[int, int, int], default to (8, 1, 1)
            The ratios of training|validation|test set.
        fields: Iterable[str], default to (User, Item)
            The fields reserved.
        """

        self.load()
        self.filter_by_rating(low=star4pos, high=None)
        self.filter_by_core(low4user=kcore4user, low4item=kcore4item, strict=strict)
        self.user2token()
        self.item2token()
        self.sort_by_timestamp()
        if fields:
            self.reserve(fields)
        self.gen_split_by_ratio(ratios)

        code = f"{kcore4user}{kcore4item}{star4pos}{''.join(map(str, ratios))}"
        path = os.path.join(
            self.root, 'General', 
            '_'.join([self.dataset, code, 'Chron'])
        )

        self.save(path)

    def make_sequential_dataset(
        self,
        star4pos: int = 0,
        kcore4user: int = 5,
        kcore4item: int = 5,
        strict: bool = True,
        fields: Optional[Iterable[str]] = (USER.name, ITEM.name, TIMESTAMP.name)
    ):
        r"""
        Make sequential dataset by leaving last two as validation|test samples.

        Parameters:
        -----------
        star4pos: int, default to 0
            Select interactions with `Rating > star4pos'.
        kcore4user: int, default to 10
            Select kcore interactions according to User.
        kcore4item: int, default to 10
            Select kcore interactions according to Item.
        strict: bool, default to `True`
            `True`: strictly filter by core
            `False`: filter by core only once
        fields: Iterable[str], default to (User, Item, TimeStamp)
            The fields reserved.
        """

        self.load()
        self.filter_by_rating(low=star4pos, high=None)
        self.filter_by_core(low4user=kcore4user, low4item=kcore4item, strict=strict)
        self.user2token()
        self.item2token()
        self.sort_by_timestamp()
        if fields:
            self.reserve(fields)
        self.seq_split_by_last_two()

        code = f"{kcore4user}{kcore4item}{star4pos}"
        path = os.path.join(
            self.root, 'Sequential', 
            '_'.join([self.dataset, code, 'Chron'])
        )

        self.save(path)

    def make_sequential_dataset_by_ratio(
        self,
        star4pos: int = 0,
        kcore4user: int = 5,
        kcore4item: int = 5,
        strict: bool = True,
        ratios: Tuple[int, int, int] = (8, 1, 1),
        fields: Optional[Iterable[str]] = (USER.name, ITEM.name, TIMESTAMP.name),
        seed: int = 0
    ):
        r"""
        Make sequential dataset by ratios.

        Parameters:
        -----------
        star4pos: int, default to 0
            Select interactions with `Rating > star4pos'.
        kcore4user: int, default to 10
            Select kcore interactions according to User.
        kcore4item: int, default to 10
            Select kcore interactions according to Item.
        strict: bool, default to `True`
            `True`: strictly filter by core
            `False`: filter by core only once
        ratios: Tuple[int, int, int], default to (8, 1, 1)
            The ratios of training|validation|test set.
        fields: Iterable[str], default to (User, Item, Timestamp)
            The fields reserved.
        """
        self.load()
        self.filter_by_rating(low=star4pos, high=None)
        self.filter_by_core(low4user=kcore4user, low4item=kcore4item, strict=strict)
        self.user2token()
        self.item2token()
        self.sort_by_timestamp()
        if fields:
            self.reserve(fields)
        self.seq_split_by_ratio(ratios, seed=seed)

        code = f"{kcore4user}{kcore4item}{star4pos}{''.join(map(str, ratios))}"
        path = os.path.join(
            self.root, 'Sequential', 
            '_'.join([self.dataset, code, 'Chron'])
        )

        self.save(path)

    def make_session_dataset_by_day(
        self,
        star4pos: int = 0,
        kcore4user: int = 2,
        kcore4item: int = 5,
        strict: bool = True,
        days: int = 7,
        fields: Optional[Iterable[str]] = (SESSION.name, ITEM.name, TIMESTAMP.name),
    ):
        r"""
        Make session dataset by day.

        Flows:
        ------
        1. filter out `inactive' items and short sessions;
        2. training|validation|test set will be splited according to the start time of each session.

        Parameters:
        -----------
        star4pos: int, default to 0
            Select interactions with `Rating > star4pos'.
        kcore4user: int, default to 10
            Select kcore interactions according to User.
        kcore4item: int, default to 10
            Select kcore interactions according to Item.
        strict: bool, default to `True`
            `True`: strictly filter by core
            `False`: filter by core only once
        days: int, default to 7
            the number days used for validation and test
        fields: Iterable[str], default to (User, Item, Timestamp)
            The fields reserved.
        """
        self.load()
        self.filter_by_rating(low=star4pos, high=None)
        self.filter_by_core(low4user=kcore4user, low4item=kcore4item, master=SESSION.name, strict=strict)
        self.user2token(master=SESSION.name)
        self.item2token()
        self.sort_by_timestamp(master=SESSION.name)
        if fields:
            self.reserve(fields)
        self.sess_split_by_day(days)

        code = f"{kcore4user}{kcore4item}{star4pos}{days}"
        path = os.path.join(
            self.root, 'Session', 
            '_'.join([self.dataset, code, 'Chron'])
        )

        self.save(path)

    def make_session_dataset_by_ratio(
        self,
        star4pos: int = 0,
        kcore4user: int = 2,
        kcore4item: int = 5,
        strict: bool = True,
        ratios: Tuple[int, int, int] = (8, 1, 1),
        fields: Optional[Iterable[str]] = (SESSION.name, ITEM.name, TIMESTAMP.name),
    ):
        r"""
        Make session dataset by ratios.

        Flows:
        ------
        1. filter out `inactive' items and short sessions;
        2. training|validation|test set will be splited according to the start time of each session.

        Parameters:
        -----------
        star4pos: int, default to 0
            Select interactions with `Rating > star4pos'.
        kcore4user: int, default to 10
            Select kcore interactions according to User.
        kcore4item: int, default to 10
            Select kcore interactions according to Item.
        strict: bool, default to `True`
            `True`: strictly filter by core
            `False`: filter by core only once
        ratios: Tuple[int, int, int], default to (8, 1, 1)
            The ratios of training|validation|test set.
        fields: Iterable[str], default to (User, Item, Timestamp)
            The fields reserved.
        """
        self.load()
        self.filter_by_rating(low=star4pos, high=None)
        self.filter_by_core(low4user=kcore4user, low4item=kcore4item, master=SESSION.name, strict=strict)
        self.user2token(master=SESSION.name)
        self.item2token()
        self.sort_by_timestamp(master=SESSION.name)
        if fields:
            self.reserve(fields)
        self.sess_split_by_ratio(ratios)

        code = f"{kcore4user}{kcore4item}{star4pos}{''.join(map(str, ratios))}"
        path = os.path.join(
            self.root, 'Session', 
            '_'.join([self.dataset, code, 'Chron'])
        )

        self.save(path)

    def summary(self):
        from prettytable import PrettyTable
        table = PrettyTable(['#User', '#Item', '#Interactions', '#Train', '#Valid', '#Test', 'Density'])
        trainsize = len(self.trainiter)
        validsize = len(self.validiter)
        testsize = len(self.testiter)
        table.add_row([
            self.userCount, self.itemCount, trainsize + validsize + testsize,
            trainsize, validsize, testsize,
            (trainsize + validsize + testsize) / (self.userCount * self.itemCount)
        ])
        infoLogger(table)




from typing import Callable, Union, Iterable

import torchdata.datapipes as dp
from ..fields import Field, FieldTuple


__all__ = ['BaseProcessor', 'Postprocessor']


class BaseProcessor(dp.iter.IterDataPipe):
    r"""
    A base processor that defines the property of fields.

    Parameters:
    -----------
    fields: Field or Iterable, optional
        - `None': Pass.
        - `Field`: FieldTuple with one Field.
        - `Iterable`: FieldTuple with multi Fields
    
    Raises:
    -------
    AttributeError: 
        If `fields' are not given or `None` before using.
    """

    def __init__(self, fields: Union[None, Field, Iterable] = None) -> None:
        super().__init__()
        self.fields = fields

    @property
    def fields(self) -> FieldTuple:
        try:
            return self.__fields
        except AttributeError:
            raise AttributeError(
                f"{self.__class__.___name__}.fields should be given (not None) before using."
            )

    @fields.setter
    def fields(self, fields: Union[None, Field, Iterable] = None):
        r"""
        Set fields.

        Parameters:
        -----------
        fields: Field or Iterable, optional
            - `None': Pass.
            - `Field`: FieldTuple with one Field.
            - `Iterable`: FieldTuple with multi Fields.
        """
        if fields is None:
            pass
        elif isinstance(fields, Field):
            self.__fields = FieldTuple((Field,))
        elif isinstance(fields, Iterable):
            self.__fields = FieldTuple(fields)
        else:
            raise ValueError(
                f"None|Field|Iterable type expected but {type(fields)} received ..."
            )
 
    @staticmethod
    def listmap(func: Callable, *iterables):
        r"""
        Apply a function to multiple iterables and return a list.

        Parameters:
        -----------
        func (Callable): The function to be applied.
        *iterables: Multiple iterables to be processed.

        Returns:
        --------
        List: The results after applying the function to the iterables.
        """
        return list(map(func, *iterables))

   

class Postprocessor(BaseProcessor):
    r"""
    A post-processor that wraps another IterDataPipe object.

    Parameters:
    -----------
    source_dp: dp.iter.IterDataPipe 
        The data pipeline to be wrapped.
    """


    def __init__(
        self, source_dp: dp.iter.IterDataPipe,
        *, 
        fields: Union[None, Field, Iterable] = None
    ) -> None:
        super().__init__(fields=fields)
        self.source = source_dp




from typing import List

import math
import numpy as np


__all__ = ['Identifier', 'Indexer', 'StandardScaler', 'MinMaxScaler']


class TransformError(Exception): ...


class Identifier:
    r"""
    Transform X into X identically.

    Parameters:
    -----------
    X : Any
        The input data to be transformed.

    Returns:
    --------
    Any
        The transformed data.

    Examples:
    ---------
    >>> col = [3, 2, 1]
    >>> transformer = Identifier()
    >>> transformer.transform(col)
    [3, 2, 1]
    """

    def __init__(self) -> None:
        self.reset()

    def reset(self): ...

    def partial_fit(self, col: List): ...

    def transform(self, col: List) -> List:
        return col


class Indexer(Identifier):
    r"""
    Transform sparse items into indices.

    classes: set
        Set of unique classes seen during `fit`.
    mapper: dict
        Mapping of input values to output indices.

    Examples:
    ---------
    >>> col = [3, 2, 1]
    >>> col2 = [4, 5, 6]
    >>> transformer = Indexer()
    >>> transformer.partial_fit(col)
    >>> transformer.classes
    {1, 2, 3}
    >>> transformer.mapper
    {1: 0, 2: 1, 3: 2}
    >>> transformer.partial_fit(col2)
    >>> transformer.classes
    {1, 2, 3, 4, 5, 6}
    >>> transformer.mapper
    {1: 0, 2: 1, 3: 2, 4: 3, 5: 4, 6: 5}
    >>> transformer.transform(col2)
    [3, 4, 5]
    """

    def reset(self):
        self.classes = set()
        self.mapper = None
        self.enums = None

    def partial_fit(self, col: List):
        self.classes |= set(col)
        ordered = sorted(self.classes)
        self.count = len(ordered)
        self.enums = tuple(range(self.count))
        self.mapper = dict(zip(ordered, self.enums))

    def _map(self, x):
        return self.mapper[x]

    def transform(self, col: List) -> List:
        if self.mapper:
            return list(map(self._map, col))
        else:
            raise TransformError("Indexer should be (partially) fitted before using ...")


class UpIndexer(Indexer):
    r"""
    Transform sparse items into indices from zero to maximum.

    Examples:
    ---------
    >>> col = [3, 2, 1]
    >>> col2 = [4, 5, 6]
    >>> transformer = UpIndexer()
    >>> transformer.partial_fit(col)
    >>> transformer.count
    4
    >>> transformer.enums
    (0, 1, 2, 3)
    >>> transformer.partial_fit(col2)
    >>> transformer.count
    7
    >>> transformer.enums
    (0, 1, 2, 3, 4, 5, 6)
    >>> transformer.transform(col2)
    [4, 5, 6]
    """

    def reset(self):
        self.enums = None
        self.count = float("-inf")

    def partial_fit(self, col: List):
        self.count = max(np.max(col).item() + 1, self.count)
        self.enums = tuple(range(self.count))

    def transform(self, col: List) -> List:
        if self.enums:
            return col
        else:
            raise TransformError("Indexer should be (partially) fitted before using ...")

class NumIndexer(Indexer):
    r"""
    Transform sparse items into indices from zero to maximum.

    Parameters:
    -----------
    nums: int
        the maximum index.

    Examples:
    ---------
    >>> transformer = NumIndexer(6)
    >>> transformer.count
    7
    >>> transformer.enums
    (0, 1, 2, 3, 4, 5, 6)
    """

    def __init__(self, nums: int) -> None:
        self.nums = nums
        super().__init__()

    def reset(self):
        self.count = self.nums + 1
        self.enums = tuple(range(self.count))

    def partial_fit(self, col: List): ...

    def transform(self, col: List) -> List:
        return col


class StandardScaler(Identifier):
    r"""
    Normalize dense items using the standard scaler.

    Attributes:
    -----------
    nums: int
        The number of fitted items.
    sum: float
        The summation of the fitted items.
    ssum: float 
        The summation of the squared fitted items.

    Examples:
    ---------
    >>> col = [3., 2., 1.]
    >>> transformer = StandardScaler()
    >>> transformer.partial_fit(col)
    >>> transformer.transform(col)
    [1.2247448713915887, 0.0, -1.2247448713915887]
    """

    def reset(self):
        self.nums = 0
        self.sum = 0
        self.ssum = 0 # sum of squared

    @property
    def mean(self):
        return self.sum / self.nums

    @property
    def std(self):
        return math.sqrt(self.ssum / self.nums - self.mean ** 2)

    def partial_fit(self, col: List):
        col = np.array(col)
        self.nums += len(col)
        self.sum += col.sum().item()
        self.ssum += (col ** 2).sum().item()

    def transform(self, col: List) -> List:
        if self.nums:
            return ((np.array(col) - self.mean) / self.std).tolist()
        else:
            raise TransformError("StandardScaler should be (partially) fitted before using ...")


class MinMaxScaler(Identifier):
    r"""
    Scale data to the range [0, 1].

    Attributes:
    -----------
    min: float
        The minimum value in the data.
    max: float
        The maximum value in the data.

    Examples:
    ---------
    >>> col = [3., 2., 1.]
    >>> transformer = MinMaxScaler()
    >>> transformer.partial_fit(col)
    >>> transformer.transform(col)
    [1.0, 0.5, 0.0]
    """

    def reset(self):
        self.min = float('inf')
        self.max = float('-inf')

    def partial_fit(self, col: List):
        self.min = min(np.min(col).item(), self.min)
        self.max = max(np.max(col).item(), self.max)

    def transform(self, col: List) -> List:
        if self.min < self.max:
            return ((np.array(col) - self.min) / (self.max - self.min)).tolist()
        else: # Same
            return col


if __name__ == "__main__":
    import doctest
    doctest.testmod()




from typing import List, Tuple, Optional, Iterable, Callable

import random
import torchdata.datapipes as dp
from collections import defaultdict

from .base import Postprocessor
from ..datasets.base import RecDataSet
from ..fields import SparseField
from ..tags import USER, SESSION, ITEM, ID
from ..utils import negsamp_vectorized_bsearch
from ...utils import timemeter


__all__ = [
    'GenTrainYielder', 'GenValidYielder', 'GenTestYielder',
    'GenTrainUniformSampler', 'GenValidYielder', 'GenTestYielder',
    'SeqTrainYielder', 'SeqValidYielder', 'SeqTestYielder', 
    'SeqTrainUniformSampler', 'SeqValidSampler', 'SeqTestSampler',
    'SessTrainYielder', 'SessValidYielder', 'SessTestYielder', 
    'SessTrainUniformSampler', 'SessValidSampler', 'SessTestSampler',
]


NUM_NEGS_FOR_SAMPLE_BASED_RANKING = 100


def _to_tuple(func: Callable):
    def wrapper(*args, **kwargs) -> Tuple:
        return tuple(func(*args, **kwargs))
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


#===============================For General Recommendation===============================


@dp.functional_datapipe("gen_train_yielding_")
class GenTrainYielder(Postprocessor):
    r"""A datapipe that yields (user, item) pairs."""

    def __init__(
        self, source_dp: dp.iter.IterDataPipe,
        dataset: RecDataSet
    ) -> None:
        r"""
        Initializes a new instance of the Tripleter postprocessor.
        
        Parameters:
        -----------
        source_dp: RecDatapipe or Postprocessor 
            A RecDatapipe or another Postprocessor that yields dicts of NumPy arrays.
        dataset: RecDataSet 
            The dataset that provides the data source.
        """
        super().__init__(source_dp)

        self.User: SparseField = dataset.fields[USER, ID]
        self.Item: SparseField = dataset.fields[ITEM, ID]

        self.prepare(dataset)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepares the dataset by building sets of seen items for each user.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset that provides the data source.
        """
        self.posItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        # sorting for ordered positives
        self.posItems = [tuple(sorted(items)) for items in self.posItems]

    def _sample_pos(self, user: int) -> int:
        r"""
        Randomly sample a positive item for a user.

        Parameters:
        -----------
        user: int 
            A user index.

        Returns:
        --------
        positive: int 
            A positive item that the user has interacted with.
        """
        return random.choice(self.posItems[user])

    def _check(self, user: int) -> bool:
        return len(self.posItems[user]) > 0

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                yield [user, self._sample_pos(user)]


@dp.functional_datapipe("gen_valid_yielding_")
class GenValidYielder(GenTrainYielder):

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = [set() for _ in range(self.User.count)]
        self.unseenItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.unseenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.seenItems = [tuple(items) for items in self.seenItems]
        self.unseenItems = [tuple(items) for items in self.unseenItems]

    def _check(self, user: int) -> bool:
        return len(self.unseenItems[user]) > 0

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                yield [user, self.unseenItems[user], self.seenItems[user]]


@dp.functional_datapipe("gen_test_yielding_")
class GenTestYielder(GenValidYielder):

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = [set() for _ in range(self.User.count)]
        self.unseenItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.test():
            self.listmap(
                lambda user, item: self.unseenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.seenItems = [tuple(items) for items in self.seenItems]
        self.unseenItems = [tuple(items) for items in self.unseenItems]


@dp.functional_datapipe("gen_train_uniform_sampling_")
class GenTrainUniformSampler(GenTrainYielder):
    r"""
    A functional datapipe for uniformly sampling users and their negatives.

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    num_negatives: int 
        The number of negative samples for each piece of data.  
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        dataset: RecDataSet,
        num_negatives: int = 1,
    ) -> None:
        self.num_negatives = num_negatives
        super().__init__(source_dp, dataset)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.posItems = [set() for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        # sorting for ordered positives
        self.posItems = [tuple(sorted(items)) for items in self.posItems]

    def _sample_neg(self, user: int) -> List[int]:
        r"""Randomly sample negative items for a user.

        Parameters:
        ----------
        user: int 
            A user index.

        Returns:
        --------
        negatives: List[int] 
            A list of negative items that the user has not interacted with.
        """
        seen = self.posItems[user]
        return negsamp_vectorized_bsearch(
            seen, self.Item.count, self.num_negatives
        )

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                yield [user, self._sample_pos(user), self._sample_neg(user)]


@dp.functional_datapipe("gen_valid_sampling_")
class GenValidSampler(GenValidYielder):

    def __init__(
        self, 
        source_dp: dp.iter.IterDataPipe, 
        dataset: RecDataSet,
        num_negs_for_sample_based_ranking: int = NUM_NEGS_FOR_SAMPLE_BASED_RANKING
    ) -> None:
        self.num_negs_for_sample_based_ranking = num_negs_for_sample_based_ranking
        super().__init__(source_dp, dataset)

    def _sample_negs(self, user: int, posItem: int):
        idx = (user, posItem)
        if self.negItems.get(idx, None) is None:
            seen = self.seenItems[user]
            self.negItems[idx] = tuple(
                negsamp_vectorized_bsearch(
                    seen, self.Item.count, self.num_negs_for_sample_based_ranking
                ).tolist()
            )
        return self.negItems[idx]

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.negItems = dict()
        # sorting for ordered positives
        self.seenItems = [tuple(sorted(items)) for items in self.seenItems]

    def __iter__(self):
        for user, posItem in self.source:
            yield [user, (posItem,) + self._sample_negs(user, posItem)]


@dp.functional_datapipe("gen_test_sampling_")
class GenTestSampler(GenValidSampler):

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.negItems = dict()
        # sorting for ordered positives
        self.seenItems = [tuple(sorted(items)) for items in self.seenItems]


#===============================For Sequential Recommendation===============================


@dp.functional_datapipe("seq_train_yielding_")
class SeqTrainYielder(Postprocessor):
    r"""
    A functional datapipe for yielding (user, positives, targets).

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    leave_one_out: bool, default to `True`
        `True`: take the last one as a target
        `False`: take `posItems[1:]` as targets
    """

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper,
        dataset: Optional[RecDataSet] = None,
        leave_one_out: bool = True
    ) -> None:
        super().__init__(source_dp)
        self.User = dataset.fields[USER, ID]
        self.Item = dataset.fields[ITEM, ID]
        if leave_one_out:
            self.marker = -1
        else:
            self.marker = 1

        self.prepare(dataset)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        pass

    def _check(self, seq: Tuple) -> bool:
        return len(seq) > 1

    def __iter__(self):
        for user, seq in self.source:
            if self._check(seq):
                yield [user, seq[:-1], seq[self.marker:]]


@dp.functional_datapipe("seq_valid_yielding_")
class SeqValidYielder(SeqTrainYielder):

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper, 
        dataset: RecDataSet
    ) -> None:
        super().__init__(source_dp, dataset, True)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]

    def __iter__(self):
        for user in self.source:
            posItems = self.posItems[user]
            if self._check(posItems):
                # (user, seqs, unseen, seen)
                yield [user, posItems[:-1], posItems[-1:], posItems[:-1]]


@dp.functional_datapipe("seq_test_yielding_")
class SeqTestYielder(SeqValidYielder):

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.test():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]


@dp.functional_datapipe("seq_train_uniform_sampling_")
class SeqTrainUniformSampler(SeqTrainYielder):
    r"""
    A functional datapipe for yielding (user, positives, targets).

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    leave_one_out: bool, default to `True`
        `True`: take the last one as a target
        `False`: take `posItems[1:]` as targets
    num_negatives: int
        The number of negatives.
    """

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper, 
        dataset: Optional[RecDataSet] = None, 
        leave_one_out: bool = True,
        num_negatives: int = 1
    ) -> None:
        self.num_negatives = num_negatives
        super().__init__(source_dp, dataset, leave_one_out)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        # sorting for ordered positives
        self.posItems = [tuple(sorted(items)) for items in self.posItems]

    def _sample_neg(self, user: int, positives: Tuple) -> List[int]:
        r"""Randomly sample negative items for a user.

        Parameters:
        ----------
        user: int
        positives: Tuple 
            A tuple of positives.

        Returns:
        --------
        negatives: List[int] 
            A list of negative items that the user has not interacted with.
        """
        seen = self.posItems[user]
        return negsamp_vectorized_bsearch(
            seen, self.Item.count, (len(positives), self.num_negatives)
        )

    def __iter__(self):
        for user, seq in self.source:
            if self._check(seq):
                seen = seq[:-1]
                positives = seq[self.marker:]
                negatives = self._sample_neg(user, positives)
                yield [user, seen, positives, negatives]


@dp.functional_datapipe("seq_valid_sampling_")
class SeqValidSampler(SeqValidYielder):

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper, 
        dataset: RecDataSet,
        num_negs_for_sample_based_ranking: int = NUM_NEGS_FOR_SAMPLE_BASED_RANKING
    ) -> None:
        self.num_negs_for_sample_based_ranking = num_negs_for_sample_based_ranking
        super().__init__(source_dp, dataset)

    @_to_tuple
    def _sample_negs(self, seen: List[int]):
        # sorting for ordered positives
        seen = sorted(seen)
        return negsamp_vectorized_bsearch(
            seen, self.Item.count, self.num_negs_for_sample_based_ranking
        ).tolist()

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]

        self.negItems = self.listmap(
            self._sample_negs, self.posItems
        )

    def __iter__(self):
        for user in self.source:
            posItems = self.posItems[user]
            if self._check(posItems):
                # (user, seq, positive || negatives)
                yield [user, posItems[:-1], posItems[-1:] + self.negItems[user]]


@dp.functional_datapipe("seq_test_sampling_")
class SeqTestSampler(SeqValidSampler):

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        for chunk in dataset.test():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]

        self.negItems = self.listmap(
            self._sample_negs, self.posItems
        )


#===============================For Session Recommendation===============================


@dp.functional_datapipe("sess_train_yielding_")
class SessTrainYielder(Postprocessor):
    r"""
    A functional datapipe for yielding (sess, sequences, targets).

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    leave_one_out: bool, default to `True`
        `True`: take the last one as a target
        `False`: take `posItems[1:]` as targets
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        dataset: Optional[RecDataSet] = None,
        leave_one_out: bool = True
    ) -> None:
        super().__init__(source_dp)
        self.Item = dataset.fields[ITEM, ID]
        if leave_one_out:
            self.marker = -1
        else:
            self.marker = 1

        self.prepare(dataset)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        pass

    def _check(self, seq) -> bool:
        return len(seq) > 1

    def __iter__(self):
        for sess, seq in self.source:
            if self._check(seq):
                yield [sess, seq[:-1], seq[self.marker:]]


@dp.functional_datapipe("sess_valid_yielding_")
class SessValidYielder(SessTrainYielder):

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper, 
        dataset: RecDataSet
    ) -> None:
        super().__init__(source_dp, dataset, True)

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = set()
        for chunk in dataset.train():
            self.listmap(
                lambda item: self.seenItems.add(item),
                chunk[ITEM, ID]
            )

    def _check(self, seq) -> bool:
        # filter out the sequence with a target not appearing in training set
        return len(seq) > 1 and seq[-1] in self.seenItems

    def __iter__(self):
        for sess, seq in self.source:
            if self._check(seq):
                # (user, seqs, unseen, seen)
                yield [sess, seq[:-1], seq[-1:], seq[:-1]]


@dp.functional_datapipe("sess_test_yielding_")
class SessTestYielder(SessValidYielder): ...


@dp.functional_datapipe("sess_train_uniform_sampling_")
class SessTrainUniformSampler(SessTrainYielder):
    r"""
    A functional datapipe for uniformly sampling negatives for each sequence.

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    num_negatives: int 
        The number of negative samples for each piece of data.  
    leave_one_out: bool, default to `True`
        `True`: take the last one as a target
        `False`: take `posItems[1:]` as targets
    num_negatives: int 
        The number of negative samples for each piece of data.  
    """

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper, 
        dataset: Optional[RecDataSet] = None, 
        leave_one_out: bool = True,
        num_negatives: int = 1
    ) -> None:
        self.num_negatives = num_negatives
        super().__init__(source_dp, dataset, leave_one_out)

    def _sample_neg(self, seen: Tuple, positives: Tuple) -> List[int]:
        r"""Randomly sample negative items for a user.

        Parameters:
        ----------
        seen: Tuple 
            A sequence of seen items.
        positives: Tuple
            A sequence of positive items.

        Returns:
        --------
        negatives: List[int] 
            A list of negative items that the user has not interacted with.
        """
        # sorting for ordered positives
        seen = sorted(seen)
        return negsamp_vectorized_bsearch(
            seen, self.Item.count, (len(positives), self.num_negatives)
        )

    def __iter__(self):
        for sess, seq in self.source:
            if self._check(seq):
                seen = seq[:-1]
                positives = seq[self.marker:]
                negatives = self._sample_neg(seq, positives)
                yield [sess, seen, positives, negatives]


@dp.functional_datapipe("sess_valid_sampling_")
class SessValidSampler(SessTrainYielder):

    def __init__(
        self, 
        source_dp: dp.iter.IterableWrapper, 
        dataset: Optional[RecDataSet] = None,
        num_negs_for_sample_based_ranking: int = NUM_NEGS_FOR_SAMPLE_BASED_RANKING
    ) -> None:
        self.num_negs_for_sample_based_ranking = num_negs_for_sample_based_ranking
        super().__init__(source_dp, dataset, True)

    def _sample_negs(self, sess: int, seq: Tuple):
        idx = (sess, tuple(seq))
        if self.negItems.get(idx, None) is None:
            seen = sorted(self.seenItems[sess])
            self.negItems[idx] = tuple(
                negsamp_vectorized_bsearch(
                    seen, self.Item.count, self.num_negs_for_sample_based_ranking
                ).tolist()
            )
        return self.negItems[idx]

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = defaultdict(set)

        for chunk in dataset.valid():
            self.listmap(
                lambda id_, item: self.seenItems[id_].add(item),
                chunk[SESSION, ID], chunk[ITEM, ID]
            )

        self.negItems = dict()
        for key in self.seenItems:
            self.seenItems[key] = tuple(self.seenItems[key])

    def __iter__(self):
        for sess, seq in self.source:
            if self._check(seq):
                seen = seq[:-1]
                posItem = seq[-1]
                yield [sess, seen, (posItem,) + self._sample_negs(sess, seq)]


@dp.functional_datapipe("sess_test_sampling_")
class SessTestSampler(SessValidSampler):

    @timemeter
    def prepare(self, dataset: RecDataSet):
        self.seenItems = defaultdict(set)

        for chunk in dataset.test():
            self.listmap(
                lambda id_, item: self.seenItems[id_].add(item),
                chunk[SESSION, ID], chunk[ITEM, ID]
            )

        self.negItems = dict()
        for key in self.seenItems:
            self.seenItems[key] = tuple(self.seenItems[key])




import torch
# import torchdata.datapipes as dp
# from torchdata.dataloader2 import DataLoader2, MultiProcessingReadingService

from .datasets import GeneralRecSet, SequentialRecSet, SessionBasedRecSet
from .postprocessing.source import OrderedIDs, OrderedSource
from .tags import USER, SESSION, ITEM, TIMESTAMP, ID, UNSEEN, SEEN, POSITIVE


__all__ = [
    'DataLoader',
    'load_gen_validpipe', 'load_gen_testpipe',
    'load_seq_lpad_validpipe', 'load_seq_lpad_testpipe',
    'load_seq_rpad_validpipe', 'load_seq_rpad_testpipe',
    'load_sess_lpad_validpipe', 'load_sess_lpad_testpipe',
    'load_sess_rpad_validpipe', 'load_sess_rpad_testpipe',
]


def _collate_fn(batch):
    return batch[0]

class DataLoader(torch.utils.data.DataLoader):

    def __init__(self, datapipe, num_workers: int = 0, **kwargs) -> None:
        super().__init__(
            dataset=datapipe, num_workers=num_workers, 
            batch_size=1, collate_fn=_collate_fn,
            **kwargs
        )


def _load_gen_datapipe(
    dataset: GeneralRecSet, mode: str,
    batch_size: int, ranking: str,
    num_negs_for_sample_based_ranking: int = 100
):
    r"""
    Load datapipe for general recommendation.

    Parameters:
    -----------
    dataset: SequentialRecSet
    mode: str, 'valid' or 'test'
    batch_size: int
    ranking: str
        `full`: datapipe for full ranking
        `pool`: datapipe for sampled-based ranking
    num_negs_for_sample_based_ranking: int
        The number of negatives for sample-based ranking.

    Raises:
    -------
    AssertionError: invalid `mode` received
    NotImplementedError: `ranking' is out of 'full' or 'pool'
    """
    
    assert mode in ('valid', 'test'), "`mode' should be 'valid' or 'test' ..."

    User = dataset.fields[USER, ID]
    Item = dataset.fields[ITEM, ID]

    if ranking == 'full':
        datapipe = OrderedIDs(
            field=User
        ).sharding_filter().__getattr__(f"gen_{mode}_yielding_")(
            dataset # return (user, unseen, seen)
        ).batch(batch_size).column_().tensor_().field_(
            User.buffer(), Item.buffer(tags=UNSEEN), Item.buffer(tags=SEEN)
        )
    elif ranking == 'pool':
        datapipe = OrderedSource(
            source=getattr(dataset, mode)().to_pairs()
        ).sharding_filter().__getattr__(f"gen_{mode}_sampling_")(
            dataset, num_negs_for_sample_based_ranking # return (user, pool)
        ).batch(batch_size).column_().tensor_()
    else:
        raise NotImplementedError(f"{ranking} ranking is not supported ...")
    
    return datapipe

def load_gen_validpipe(
    dataset: SequentialRecSet,
    batch_size: int = 512, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_gen_datapipe(
        dataset, mode='valid', 
        batch_size=batch_size, ranking=ranking,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_gen_testpipe(
    dataset: SequentialRecSet,
    batch_size: int = 512, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_gen_datapipe(
        dataset, mode='test', 
        batch_size=batch_size, ranking=ranking,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )


def _load_seq_datapipe(
    dataset: SequentialRecSet, mode: str,
    maxlen: int, batch_size: int, ranking: str,
    padding_way, NUM_PADS: int, padding_value: int,
    num_negs_for_sample_based_ranking: int = 100
):
    r"""
    Load datapipe for sequential recommendation.

    Parameters:
    -----------
    dataset: SequentialRecSet
    mode: str, 'valid' or 'test'
    maxlen: int
    batch_size: int
    ranking: str
        `full`: datapipe for full ranking
        `pool`: datapipe for sampled-based ranking
    padding_way: str
        `left`: sequence will be padded from left
    NUM_PADS: int
    padding_value: int
    num_negs_for_sample_based_ranking: int
        The number of negatives for sample-based ranking.

    Raises:
    -------
    AssertionError: invalid `mode` or `padding_way` received
    NotImplementedError: `ranking' is out of 'full' or 'pool'
    """
    
    assert mode in ('valid', 'test'), "`mode' should be 'valid' or 'test' ..."
    assert padding_way in ('left', 'right'), "`padding_way` should be 'left' or 'right' ..."

    padding_way = 'lpad_' if padding_way == 'left' else 'rpad_'
    User = dataset.fields[USER, ID]
    Item = dataset.fields[ITEM, ID]

    if ranking == 'full':
        datapipe = OrderedIDs(
            field=User
        ).sharding_filter().__getattr__(f"seq_{mode}_yielding_")(
            dataset # yielding (user, seq, unseen, seen)
        ).lprune_(
            indices=[1], maxlen=maxlen,
        ).add_(
            indices=[1], offset=NUM_PADS
        ).__getattr__(padding_way)(
            indices=[1], maxlen=maxlen, padding_value=padding_value
        ).batch(batch_size).column_().tensor_().field_(
            User.buffer(), Item.buffer(tags=POSITIVE), Item.buffer(tags=UNSEEN), Item.buffer(tags=SEEN)
        )
    elif ranking == 'pool':
        datapipe = OrderedIDs(
            field=User
        ).sharding_filter().__getattr__(f"seq_{mode}_sampling_")(
            dataset, num_negs_for_sample_based_ranking # yielding (user, seq, (target + (100) negatives))
        ).lprune_(
            indices=[1], maxlen=maxlen,
        ).add_(
            indices=[1, 2], offset=NUM_PADS
        ).__getattr__(padding_way)(
            indices=[1], maxlen=maxlen, padding_value=0
        ).batch(batch_size).column_().tensor_()
    else:
        raise NotImplementedError(f"{ranking} ranking is not supported ...")
    
    return datapipe

def load_seq_lpad_validpipe(
    dataset: SequentialRecSet,
    maxlen: int, NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 128, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_seq_datapipe(
        dataset, mode='valid', 
        maxlen=maxlen, batch_size=batch_size, ranking=ranking, 
        padding_way='left', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_seq_rpad_validpipe(
    dataset: SequentialRecSet,
    maxlen: int, NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 128, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_seq_datapipe(
        dataset, mode='valid', 
        maxlen=maxlen, batch_size=batch_size, ranking=ranking, 
        padding_way='right', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_seq_lpad_testpipe(
    dataset: SequentialRecSet,
    maxlen: int, NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 128, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_seq_datapipe(
        dataset, mode='test', 
        maxlen=maxlen, batch_size=batch_size, ranking=ranking, 
        padding_way='left', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_seq_rpad_testpipe(
    dataset: SequentialRecSet,
    maxlen: int, NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 128, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_seq_datapipe(
        dataset, mode='test', 
        maxlen=maxlen, batch_size=batch_size, ranking=ranking, 
        padding_way='right', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def _load_sess_datapipe(
    dataset: SessionBasedRecSet, mode: str,
    batch_size: int, ranking: str,
    padding_way, NUM_PADS: int, padding_value: int,
    num_negs_for_sample_based_ranking: int = 100
):
    r"""
    Load datapipe for session-based recommendation.

    Parameters:
    -----------
    dataset: SessionBasedRecSet
    mode: str, 'valid' or 'test'
    batch_size: int
    ranking: str
        `full`: datapipe for full ranking
        `pool`: datapipe for sampled-based ranking
    padding_way: str
        `left`: sequence will be padded from left
    NUM_PADS: int
    padding_value: int
    num_negs_for_sample_based_ranking: int
        The number of negatives for sample-based ranking

    Raises:
    -------
    AssertionError: invalid `mode` or `padding_way` received
    NotImplementedError: `ranking' is out of 'full' or 'pool'
    """
    
    assert mode in ('valid', 'test'), "`mode' should be 'valid' or 'test' ..."
    assert padding_way in ('left', 'right'), "`padding_way` should be 'left' or 'right' ..."

    padding_way = 'lpad_' if padding_way == 'left' else 'rpad_'
    Session = dataset.fields[SESSION, ID]
    Item = dataset.fields[ITEM, ID]

    if ranking == 'full':
        datapipe = OrderedSource(
            source=getattr(dataset, mode)().to_roll_seqs(minlen=2)
        ).sharding_filter().__getattr__(f"sess_{mode}_yielding_")(
            dataset # yielding (sesses, seq, unseen, seen)
        ).add_(
            indices=[1], offset=NUM_PADS
        ).batch(batch_size).column_().__getattr__(f"{padding_way}col_")(
            indices=[1], maxlen=None, padding_value=padding_value
        ).tensor_().field_(
            Session.buffer(), Item.buffer(tags=POSITIVE), Item.buffer(tags=UNSEEN), Item.buffer(tags=SEEN)
        )
    elif ranking == 'pool':
        datapipe = OrderedSource(
            source=getattr(dataset, mode)().to_roll_seqs(minlen=2)
        ).sharding_filter().__getattr__(f"sess_{mode}_sampling_")(
            dataset, num_negs_for_sample_based_ranking # yielding (sesses, seq, (target + (100) negatives))
        ).add_(
            indices=[1, 2], offset=NUM_PADS
        ).batch(batch_size).column_().__getattr__(f"{padding_way}col_")(
            indices=[1], maxlen=None, padding_value=padding_value
        ).tensor_()
    else:
        raise NotImplementedError(f"{ranking} ranking is not supported ...")
    
    return datapipe

def load_sess_lpad_validpipe(
    dataset: SessionBasedRecSet,
    NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 256, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_sess_datapipe(
        dataset, mode='valid', 
        batch_size=batch_size, ranking=ranking, 
        padding_way='left', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_sess_rpad_validpipe(
    dataset: SessionBasedRecSet,
    NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 256, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_sess_datapipe(
        dataset, mode='valid', 
        batch_size=batch_size, ranking=ranking, 
        padding_way='right', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_sess_lpad_testpipe(
    dataset: SessionBasedRecSet,
    NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 256, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_sess_datapipe(
        dataset, mode='test', 
        batch_size=batch_size, ranking=ranking, 
        padding_way='left', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )

def load_sess_rpad_testpipe(
    dataset: SessionBasedRecSet,
    NUM_PADS: int, padding_value: int = 0,
    batch_size: int = 256, ranking: str = 'full',
    num_negs_for_sample_based_ranking: int = 100
):
    return _load_sess_datapipe(
        dataset, mode='test', 
        batch_size=batch_size, ranking=ranking, 
        padding_way='right', NUM_PADS=NUM_PADS, padding_value=padding_value,
        num_negs_for_sample_based_ranking=num_negs_for_sample_based_ranking
    )




from typing import Any, Set, Union, List

import torch
import os, argparse, time, yaml
from argparse import ArgumentParser

from .dict2obj import Config
from .ddp import is_distributed, main_process_only, is_main_process, all_gather
from .utils import (
    mkdirs, timemeter, set_color, set_seed, activate_benchmark, 
    set_logger, infoLogger
)


__all__ = ['Parser', 'CoreParser']


r"""
A configuration module for a Recommender System.

Templates:
----------
DATA_DIR : str
    Directory for saving historical metric information.
SUMMARY_DIR : str
    Directory for saving final summary information.
CHECKPOINT_PATH : str
    Path for saving checkpoints.
LOG_PATH : str
    Path for saving logging information.
CORE_CHECKPOINT_PATH : str
    Path for tuning.
CORE_LOG_PATH : str
    Path for tuning logging.
TIME : str
    A string representation of time in the format of "%m%d%H%M%S".
CONFIG : Config
    A Config object containing general configurations for the recommender system.
CORE_CONFIG : Config
    A Config object containing configurations for tuning the recommender system.


Configurations:
---------------
SAVED_FILENAME : str
    The filename of saved model parameters.
BEST_FILENAME : str
    The filename of saved best model parameters.
CHECKPOINT_FREQ : int
    The frequency of saving checkpoints.
CHECKPOINT_MODULES : list
    A list of modules to save in a checkpoint.
CHECKPOINT_FILENAME : str
    The filename of the saved checkpoint.
SUMMARY_FILENAME : str
    The filename of the summary.
MONITOR_FILENAME : str
    The filename of the monitor.
MONITOR_BEST_FILENAME : str
    The filename of the best monitor.
description : str
    The description of the recommender system.
EXCLUSIVE : bool
    If True, run a command exclusively.
COMMAND : str
    The command to be run.
ENVS : dict
    A dictionary of environment variables.
PARAMS : dict
    A dictionary of parameters.
DEFAULTS : dict
    A dictionary of default parameters.
"""

DATA_DIR = 'data'
SUMMARY_DIR = 'summary'
CHECKPOINT_PATH = "./infos/{description}/{dataset}/{device}"
LOG_PATH = "./logs/{description}/{dataset}/{id}"
CORE_CHECKPOINT_PATH = "./infos/{description}/core"
CORE_LOG_PATH = "./logs/{description}/core"
TIME = "%m%d%H%M%S"

CONFIG = Config(
    # path|file
    SAVED_FILENAME = "model.pt",
    BEST_FILENAME = "best.pt",
    CHECKPOINT_FREQ = 1,
    CHECKPOINT_MODULES = ['model', 'optimizer', 'lr_scheduler'],
    CHECKPOINT_FILENAME = "checkpoint.tar",
    SUMMARY_FILENAME = "SUMMARY.md",
    MONITOR_FILENAME = "monitors.pkl",
    MONITOR_BEST_FILENAME = "best.pkl",
)

CORE_CONFIG = Config(
    MONITOR_BEST_FILENAME = CONFIG.MONITOR_BEST_FILENAME,
    CHECKPOINT_FILENAME = CONFIG.CHECKPOINT_FILENAME,
    EXCLUSIVE = False,
    COMMAND = None,
    ENVS = dict(),
    PARAMS = dict(),
    DEFAULTS = dict(),

    log2file = True,
    log2console = True
)


class Parser(Config):
    """ArgumentParser wrapper."""

    # FILE
    SAVED_FILENAME: str
    BEST_FILENAME: str
    CHECKPOINT_FREQ: int
    CHECKPOINT_MODULES: List[str]
    CHECKPOINT_FILENAME: str
    SUMMARY_FILENAME: str
    MONITOR_FILENAME: str
    MONITOR_BEST_FILENAME: str

    # PATH
    DATA_DIR: str
    SUMMARY_DIR: str
    CHECKPOINT_PATH: str
    LOG_PATH: str

    def __init__(self) -> None:
        super().__init__(**CONFIG)
        self.parse()

    @main_process_only
    def readme(self, path: str, mode: str = "w") -> None:
        """Add README.md to the path."""
        time_ = time.strftime("%Y-%m-%d-%H:%M:%S")
        file_ = os.path.join(path, "README.md")
        s = "|  {key}  |   {val}    |\n"
        info = "\n## {0} \n\n\n".format(time_)
        info += "|  Attribute   |   Value   |\n"
        info += "| :-------------: | :-----------: |\n"
        for key in sorted(self.keys(), key=lambda x: x.upper()):
            info += s.format(key=key, val=self[key])
        with open(file_, mode, encoding="utf8") as fh:
            fh.write(info)

    def reset(self):
        self.clear()
        for key, val in CONFIG.items():
            self[key] = val

    @timemeter
    def parse(self):
        """Add command-line arguments to the parser."""

        self.parser = argparse.ArgumentParser()

        self.add_argument("--root", type=str, default=".", help="data path")
        self.add_argument("--dataset", type=str, default="RecDataSet", help="useless if no need to automatically select a dataset")
        self.add_argument("--config", type=str, default=None, help="config.yaml")
        self.add_argument("--ranking", type=str, choices=('full', 'pool'), default='full', help="full: full ranking; pool: sampled-based ranking")
        self.add_argument("--retain-seen", action="store_true", default=False, help="True: retain seen candidates during evaluation")

        self.add_argument("--device", default=torch.cuda.current_device() if torch.cuda.is_available() else 'cpu', help="device")
        self.add_argument("--ddp-backend", type=str, default='nccl', help="ddp backend")

        # model
        self.add_argument("--optimizer", type=str, default="adam", help="Optimizer: adam (default), sgd, ...")
        self.add_argument("--nesterov", action="store_true", default=False, help="nesterov for SGD")
        self.add_argument("-mom", "--momentum", type=float, default=0.9, help="the momentum used for SGD")
        self.add_argument("-beta1", "--beta1", type=float, default=0.9, help="the first beta argument for Adam")
        self.add_argument("-beta2", "--beta2", type=float, default=0.999, help="the second beta argument for Adam")
        self.add_argument("-wd", "--weight-decay", type=float, default=None, help="weight for 'l1|l2|...' regularzation")
        self.add_argument("-lr", "--lr", "--LR", "--learning-rate", type=float, default=None)
        self.add_argument("-b", "--batch-size", type=int, default=None)
        self.add_argument("--epochs", type=int, default=None)

        # logging
        self.add_argument("--log2file", action="store_false", default=True, help="if True, save logs to a file")
        self.add_argument("--log2console", action="store_false", default=True, help="if True, display logs on the console")

        # eval
        self.add_argument("--eval-valid", action="store_false", default=True, help="if True, evaluate validset")
        self.add_argument("--eval-test", action="store_true", default=False, help="if True, evaluate testset")
        self.add_argument("--eval-freq", type=int, default=5, help="the evaluation frequency")

        self.add_argument("--num-workers", type=int, default=4)
        self.add_argument("--pin-memory", action="store_true", default=False)

        self.add_argument("--seed", type=int, default=1, help="calling --seed=-1 for a random seed")
        self.add_argument("--benchmark", action="store_true", default=False, help="cudnn.benchmark == True ?")
        self.add_argument("--resume", action="store_true", default=False, help="resume the training from the recent checkpoint")

        self.add_argument("--id", type=str, default=time.strftime(TIME))
        self.add_argument("-m", "--description", type=str, default="RecSys")

    def add_argument(self, *args: str, **kwargs):
        r"""
        Add an argument to the parser.

        Parameters:
        -----------
        *args : str
            The name(s) of the argument.
        **kwargs
            Additional keyword arguments to pass to `parser.add_argument`.
       
        Notes:
        ------
        Any character `_' will be replaced by `-'.
        """
        args = (arg.replace('_', '-') for arg in args) # user '-' instead of '_'
        action = self.parser.add_argument(*args, **kwargs)
        self[action.dest] = action.default

    def set_defaults(self, **kwargs):
        r"""
        Set the default values for the arguments.

        Parameters:
        -----------
        **kwargs
            The default values to set.
        """
        self.parser.set_defaults(**kwargs)
        for key, val in kwargs.items():
            self[key] = val

    def set_device(self, device: Union[torch.device, str, int]):
        try:
            device = int(device)
        except ValueError:
            pass

        set_color(device)
        self.device = torch.device(device)
        try:
            torch.cuda.set_device(device)
        except ValueError:
            pass
        except AttributeError:
            pass
        return device

    def init_ddp(self):
        import torch.distributed as dist
        if is_distributed():
            dist.init_process_group(backend=self.ddp_backend, init_method="env://")
            self.device = int(os.environ["LOCAL_RANK"])
            # synchronize ids
            self.id = all_gather(self.id)[0]
            infoLogger(f"[DDP] >>> DDP is activated ...")
    
    @timemeter
    def load(self):
        r"""
        Load config.yaml.

        Raises:
        -------
        KeyError
            If the parameter key is not recognized.
        
        Returns:
        --------
        ArgParser:
            Some of defaults are overwrited by the config file
        """
        args = self.parser.parse_args()
        keys, _ = zip(*args._get_kwargs())
        if hasattr(args, 'config') and args.config:
            with open(args.config, encoding="UTF-8", mode='r') as f:
                for key, val in yaml.full_load(f).items():
                    if key in keys: # overwriting defaults
                        self.set_defaults(**{key: val})
                    elif key.upper() in self:
                        self[key.upper()] = val
                    else:
                        raise KeyError(f"Unexpected parameter of `{key}' in `{args.config}' ...")
        return self.parser.parse_args()

    @timemeter
    def compile(self):
        r"""
        Generate the configuration file according to the specified settings.

        Flows:
        ------
        1. If the `--config` flag has been specified, load settings from a .yaml file.
        2. Generate the paths for saving checkpoints and collecting training information:
            - CHECKPOINT_PATH: the path where checkpoints will be saved.
            - LOG_PATH: the path where training information will be collected.
        3. Configure the logger so that information can be logged by `info|debug|warnLogger`.
        4. Add a README.md file under CHECKPOINT_PATH and LOG_PATH.
        """
        args = self.load()
        self.update(**dict(args._get_kwargs()))
        
        self['DATA_DIR'] = DATA_DIR
        self['SUMMARY_DIR'] = SUMMARY_DIR

        self.init_ddp()
        self['CHECKPOINT_PATH'] = CHECKPOINT_PATH.format(**self)
        self['LOG_PATH'] = LOG_PATH.format(**self)
        mkdirs(
            self.CHECKPOINT_PATH, self.LOG_PATH,
            os.path.join(self.LOG_PATH, self.DATA_DIR),
            os.path.join(self.LOG_PATH, self.SUMMARY_DIR)
        )

        self.set_device(self.device)
        set_logger(path=self.LOG_PATH, log2file=self.log2file, log2console=self.log2console)

        activate_benchmark(self.benchmark)
        self.seed = set_seed(self.seed)

        self.readme(self.CHECKPOINT_PATH) # create README.md
        self.readme(self.LOG_PATH)

        if is_main_process():
            infoLogger(str(self))


class CoreParser(Config):
    """CoreParser class to parse command-line arguments and configuration files."""

    ALL_ENVS = (
        'description', 'root', 'device', 'num_workers'
    )

    def __init__(self) -> None:
        super().__init__(**CORE_CONFIG)

    def check(self) -> None:
        """Check the validity of the given config."""
        template = """
        Please make sure the configuration file follows the template below:

        command: python xxx.py
        envs:
            root: ../../data
            dataset: Gowalla_m1
            device: 0,1,2,3
            eval_freq: 5
            num_workers: 0
        params:
            optimizer: [adam, sgd]
            learning_rate: [1.e-3, 1.e-2, 1.e-1]
            weight_decay: [0, 1.e-4, 2.e-4, 5.e-4]
            batch_size: [128, 256, 512, 1024]
        defaults:
            optimizer: adam
            learning_rate: 1.e-3
            weight_decay: 0
            batch_size: 256
            epochs: 100
            seed: 1

        where 'command' is necessary but 'envs', 'params' and 'defaults' are optional ...

        Notes: when calling '--exclusive' for grid search one by one, 
            'defaults' is required for clear comparsions in tensorbaord.
        """
        if self.COMMAND is None:
            raise ValueError(template)
        for key in ('root', 'dataset', 'device'):
            if self.ENVS.get(key, None) is None:
                raise KeyError(f"No `{key}' is allocated, calling '--{key}' to specify it")

        self.ENVS = Config(self.ENVS)
        self.PARAMS = Config(self.PARAMS)

    @timemeter
    def load(self, args: ArgumentParser) -> None:
        r"""
        Load configuration file.

        Parameters:
        -----------
        args : ArgumentParser
            The command line arguments.

        Returns:
        --------
        None
        """
        with open(args.config, encoding="UTF-8", mode='r') as f:
            config = {key.upper(): vals for key, vals in yaml.full_load(f).items()}
            self.update(**config)
        self.EXCLUSIVE = args.exclusive
        self.resume = args.resume
        for key, val in args._get_kwargs():
            if key in self.ALL_ENVS and val is not None:
                self.ENVS[key] = val

    @timemeter
    def compile(self, args) -> None:
        r"""
        Generate config file according to settings.

        Flows:
        ------
        1. Load settings from xxx.yaml which provides parameters for grid searching.
        2. Load settings of the execution environment from parsed args (ArgumentParser).
        3. Generate basic paths:
            - CHECKPOINT_PATH: subprocess
            - LOG_PATH: subprocess
            - CORE_CHECKPOINT_PATH: saving checkpoints of the rest of params
            - CORE_LOG_PATH: saving best results of each subprocess for comparison
        4. Set Logger, and then you can log information by info|debug|warnLogger ...
        5. Finally, READMD.md will be added under CHECKPOINT_PATH and LOG_PATH both.
        """
        self.load(args)
        self.check()

        self['DATA_DIR'] = DATA_DIR
        self['SUMMARY_DIR'] = SUMMARY_DIR
        self['CHECKPOINT_PATH'] = CHECKPOINT_PATH
        self['LOG_PATH'] = LOG_PATH
        self['CORE_CHECKPOINT_PATH'] = CORE_CHECKPOINT_PATH.format(**self.ENVS)
        self['CORE_LOG_PATH'] = CORE_LOG_PATH.format(**self.ENVS)
        mkdirs(self.CORE_CHECKPOINT_PATH, self.CORE_LOG_PATH)
        set_logger(path=self.CORE_LOG_PATH, log2file=self.log2file, log2console=self.log2console)

        self.readme(self.CORE_LOG_PATH)

    def readme(self, path: str, mode: str = "w") -> None:
        r"""
        Add README.md to the given path.

        Parameters:
        -----------
        path: str 
            The path to add the README.md file.
        mode: str, optional 
            The file opening mode. Defaults to "w".

        Returns:
        --------
        None
        """
        time_ = time.strftime("%Y-%m-%d-%H:%M:%S")
        file_ = os.path.join(path, "README.md")
        s = "|  {key}  |   {val}    |\n"
        info = "\n## {0} \n\n\n".format(time_)
        info += "|  Attribute   |   Value   |\n"
        info += "| :-------------: | :-----------: |\n"
        for key, val in self.ENVS.items():
            info += s.format(key=key, val=val)
        for key, val in self.PARAMS.items():
            info += s.format(key=key, val=val)
        for key, val in self.DEFAULTS.items():
            info += s.format(key=f"{key} (default)", val=val)
        with open(file_, mode, encoding="utf8") as fh:
            fh.write(info)




from typing import Iterable

import numpy as np
from math import floor, ceil

from .base import AtomicConverter
from ..tags import USER, SESSION, ITEM, RATING, TIMESTAMP
from ...utils import infoLogger

class AmazonBeauty(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	title:token	    sales_type:token	sales_rank:float	categories:token_seq	price:float 	brand:token
    """
    filename = "Amazon_Beauty"


class AmazonBooks(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	sales_type:token	sales_rank:float	categories:token_seq	title:token     price:float    brand:token
    """
    filename = "Amazon_Books"


class AmazonCDs(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	title:token	    categories:token_seq	brand:token	sales_type:token	sales_rank:float	price:float
    """
    filename = "Amazon_CDs_and_Vinyl\Amazon_CDs_and_Vinyl"


class AmazonElectronics(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	categories:token_seq	title:token 	price:float	    sales_type:token	sales_rank:float	brand:token
    """
    filename = "Amazon_Electronics"


class AmazonHome(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	sales_type:token	sales_rank:float	categories:token_seq	title:token	price:float	brand:token
    """
    filename = "Amazon_Home_and_Kitchen"


class AmazonMovies(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	categories:token_seq	title:token	    price:float	    sales_type:token	sales_rank:float	brand:token
    """
    filename = "Amazon_Movies_and_TV"


class AmazonOffice(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	price:float	sales_type:token	sales_rank:float	categories:token_seq	title:token	brand:token
    """
    filename = "Amazon_Office_Products"


class AmazonTools(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	categories:token_seq	title:token	price:float	brand:token	sales_type:token	sales_rank:float
    """
    filename = "Amazon_Tools_and_Home_Improvement"


class AmazonToys(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	title:token	price:float	sales_type:token	sales_rank:float	brand:token	categories:token_seq
    """
    filename = "Amazon_Toys_and_Games"


class AmazonGames(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	rating:float	timestamp:float
    item:
        item_id:token	price:float	sales_type:token	sales_rank:float	categories:token_seq	title:token	brand:token
    """
    filename = "Amazon_Video_Games"


class Diginetica(AtomicConverter):
    r"""
    inter:
        session_id:token	item_id:token	timestamp:float
    item:
        item_id:token	item_priceLog2:float	item_name:token	item_category:token
    """
    filename = "diginetica"


class FourSquareNYC(AtomicConverter):
    r"""
    inter:
        user_id:token	venue_id:token	timezone_offset:float	timestamp:float	    click_times:float
    item:
        venue_id:token	venue_category_id:token	    venue_category_name:token	latitude:float	  longitude:float
    """
    filename = "foursquare_NYC"
    _name_format_dict = {'venue_id': ITEM.name}


class FourSquareTKY(AtomicConverter):
    r"""
    inter:
        user_id:token	venue_id:token	timezone_offset:float	timestamp:float	    click_times:float
    item:
        venue_id:token	venue_category_id:token	    venue_category_name:token	latitude:float	  longitude:float
    """
    filename = "foursquare_TKY"


class Gowalla(AtomicConverter):
    r"""
    inter:
        user_id:token	item_id:token	timestamp:float	    latitude:float	longitude:float	    num_repeat:float
    """
    filename = "gowalla"
    _name_format_dict = {'venue_id': ITEM.name}


class LastFM(AtomicConverter):
    r"""
    inter:
        user_id:token	artist_id:token	    weight:float	tag_value:token_seq
    item:
        artist_id:token	    name:token	    url:token	    picture_url:token
    """
    filename = "lastfm"


class MovieLens1M(AtomicConverter):
    r"""
    inter: 
        user_id:token	item_id:token	rating:float	timestamp:float
    user: 
        user_id:token	age:token	gender:token	occupation:token	zip_code:token
    item:
        item_id:token	movie_title:token_seq	release_year:token	class:token_seq
    """
    filename = "ml-1m"


class MovieLens10M(AtomicConverter):
    r"""
    inter: 
        user_id:token	item_id:token	rating:float	timestamp:float
    user: 
        user_id:token	age:token	gender:token	occupation:token	zip_code:token
    item:
        item_id:token	movie_title:token_seq	release_year:token	class:token_seq
    """
    filename = "ml-10m"


class MovieLens100K(AtomicConverter):
    r"""
    inter: 
        user_id:token	item_id:token	rating:float	timestamp:float
    user: 
        user_id:token	age:token	gender:token	occupation:token	zip_code:token
    item:
        item_id:token	movie_title:token_seq	release_year:token	class:token_seq
    """
    filename = "ml-100K"


class Steam(AtomicConverter):
    r"""
    inter: 
        user_id:token	play_hours:float	products:float	    product_id:token	page_order:float	timestamp:float	    early_access:token	    page:float	    found_funny:token	compensation:token	    times:float
    item:
        app_name:token	developer:token	early_access:token	genres:token_seq	id:token	metascore:float	    price:float	    publisher:token	    timestamp:float	    sentiment:token	    specs:token_seq	    tags:token_seq	    title:token
    """
    filename = "steam\steam"
    _name_format_dict = {'product_id': ITEM.name, 'id': ITEM.name}


class TmallBuy(AtomicConverter):
    r"""
    inter: 
        user_id:token	seller_id:token 	item_id:token	category_id:token	timestamp:float	    interactions:float
    """
    filename = "tmall-buy"


class TmallClick(AtomicConverter):
    r"""
    inter: 
        user_id:token	seller_id:token 	item_id:token	category_id:token	timestamp:float	    interactions:float
    """
    filename = "tmall-click"


class Yelp2018(AtomicConverter):
    r"""
    inter: 
        user_id:token	item_id:token	rating:float	timestamp:float	    useful:float	funny:float	cool:float	review_id:token
    user:
        user_id:token	user_name:token	    user_review_count:float	    yelping_since:float	    user_useful:float	user_funny:float	user_cool:float	    elite:token	    fans:float	average_stars:float
    item:
        item_id:token	item_name:token_seq	    address:token_seq	city:token_seq	    state:token	postal_code:token	latitude:float	longitude:float	    item_stars:float	item_review_count:float	is_open:float	categories:token_seq
    """
    filename = "yelp2018\yelp2018"


class Yelp2021(AtomicConverter):
    r"""
    inter: 
        user_id:token	item_id:token	rating:float	timestamp:float	    useful:float	funny:float	cool:float	review_id:token
    user:
        user_id:token	user_name:token	    user_review_count:float	    yelping_since:float	    user_useful:float	user_funny:float	user_cool:float	    elite:token	    fans:float	average_stars:float
    item:
        item_id:token	item_name:token_seq	    address:token_seq	city:token_seq	    state:token	postal_code:token	latitude:float	longitude:float	    item_stars:float	item_review_count:float	is_open:float	categories:token_seq
    """
    filename = "yelp2021\yelp2021"


class Yelp2022(AtomicConverter):
    r"""
    inter: 
        user_id:token	item_id:token	rating:float	timestamp:float	    useful:float	funny:float	cool:float	review_id:token
    user:
        user_id:token	user_name:token	    user_review_count:float	    yelping_since:float	    user_useful:float	user_funny:float	user_cool:float	    elite:token	    fans:float	average_stars:float
    item:
        item_id:token	item_name:token_seq	    address:token_seq	city:token_seq	    state:token	postal_code:token	latitude:float	longitude:float	    item_stars:float	item_review_count:float	is_open:float	categories:token_seq
    """
    filename = "yelp2022\yelp2022"


class YooChooseBuys(AtomicConverter):
    r"""
    inter: 
        session_id:token	timestamp:float	item_id:token	price:float	quantity:float
    """
    filename = "yoochoose-buys"

class YooChooseBuys14(YooChooseBuys):

    def sess_split_by_ratio(self, ratios: Iterable = (8, 1, 1)):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")

        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)
        groups = groups[-ceil(len(groups) / 4):] # recent 1/4 sessions

        markers = np.cumsum(ratios)
        l = max(floor(markers[0] * len(groups) / markers[-1]), 1)
        r = floor(markers[1] * len(groups) / markers[-1])

        traingroups = groups[:l]
        validgroups = groups[l:r]
        testgroups = groups[r:]

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]


class YooChooseBuys164(YooChooseBuys):

    def sess_split_by_ratio(self, ratios: Iterable = (8, 1, 1)):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")

        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)
        groups = groups[-ceil(len(groups) / 64):] # recent 1/64 sessions

        markers = np.cumsum(ratios)
        l = max(floor(markers[0] * len(groups) / markers[-1]), 1)
        r = floor(markers[1] * len(groups) / markers[-1])

        traingroups = groups[:l]
        validgroups = groups[l:r]
        testgroups = groups[r:]

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]


class YooChooseClicks(AtomicConverter):
    r"""
    inter: 
        session_id:token	timestamp:float	item_id:token	category:token
    """
    filename = "yoochoose-clicks"


class YooChooseClicks14(YooChooseClicks):

    def sess_split_by_ratio(self, ratios: Iterable = (8, 1, 1)):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")

        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)
        groups = groups[-ceil(len(groups) / 4):] # recent 1/4 sessions

        markers = np.cumsum(ratios)
        l = max(floor(markers[0] * len(groups) / markers[-1]), 1)
        r = floor(markers[1] * len(groups) / markers[-1])

        traingroups = groups[:l]
        validgroups = groups[l:r]
        testgroups = groups[r:]

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]

    def sess_split_by_day(self, days: int = 1):
        infoLogger(f"[Converter] >>> Split by days: {days} ...")

        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)
        groups = groups[-ceil(len(groups) / 4):] # recent 1/4 sessions
        self.interactions = self.interactions[self.interactions[SESSION.name].isin(groups)]

        seconds_per_day = 86400
        seconds = seconds_per_day * days

        last_date = self.interactions[TIMESTAMP.name].max().item()

        # Group interactions by session and calculate session timestamps
        session_timestamps = self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values()

        # Split interactions into train, validation, and test sets based on session timestamps
        traingroups = session_timestamps[session_timestamps < (last_date - 2 * seconds)].index
        validgroups = session_timestamps[(session_timestamps >= (last_date - 2 * seconds)) & (session_timestamps < (last_date - seconds))].index
        testgroups = session_timestamps[session_timestamps >= (last_date - seconds)].index

        assert len(traingroups) >= 0, f"The given `days` of {days} leads to zero-size trainsets ..."
        assert len(validgroups) >= 0, f"The given `days` of {days} leads to zero-size validsets ..."
        assert len(testgroups) >= 0, f"The given `days` of {days} leads to zero-size testsets ..."

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]


class YooChooseClicks164(YooChooseClicks):

    def sess_split_by_ratio(self, ratios: Iterable = (8, 1, 1)):
        infoLogger(f"[Converter] >>> Split by ratios: {ratios} ...")

        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)
        groups = groups[-ceil(len(groups) / 64):] # recent 1/64 sessions

        markers = np.cumsum(ratios)
        l = max(floor(markers[0] * len(groups) / markers[-1]), 1)
        r = floor(markers[1] * len(groups) / markers[-1])

        traingroups = groups[:l]
        validgroups = groups[l:r]
        testgroups = groups[r:]

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]

    def sess_split_by_day(self, days: int = 1):
        infoLogger(f"[Converter] >>> Split by days: {days} ...")

        groups = list(self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values().index)
        groups = groups[-ceil(len(groups) / 64):] # recent 1/64 sessions
        self.interactions = self.interactions[self.interactions[SESSION.name].isin(groups)]

        seconds_per_day = 86400
        seconds = seconds_per_day * days

        last_date = self.interactions[TIMESTAMP.name].max().item()

        # Group interactions by session and calculate session timestamps
        session_timestamps = self.interactions.groupby(SESSION.name)[TIMESTAMP.name].max().sort_values()

        # Split interactions into train, validation, and test sets based on session timestamps
        traingroups = session_timestamps[session_timestamps < (last_date - 2 * seconds)].index
        validgroups = session_timestamps[(session_timestamps >= (last_date - 2 * seconds)) & (session_timestamps < (last_date - seconds))].index
        testgroups = session_timestamps[session_timestamps >= (last_date - seconds)].index

        assert len(traingroups) >= 0, f"The given `days` of {days} leads to zero-size trainsets ..."
        assert len(validgroups) >= 0, f"The given `days` of {days} leads to zero-size validsets ..."
        assert len(testgroups) >= 0, f"The given `days` of {days} leads to zero-size testsets ..."

        self.trainiter = self.interactions[self.interactions[SESSION.name].isin(traingroups)]
        self.validiter = self.interactions[self.interactions[SESSION.name].isin(validgroups)]
        self.testiter = self.interactions[self.interactions[SESSION.name].isin(testgroups)]

